# -*- coding: utf-8 -*-

"""
This module hosts common utilities for Simmate's workflow library. This includes
functions for grabbing all available workflows as well as dynamically loading
a workflow using its name.
"""

import importlib
import logging
import shutil
import sys
from inspect import getmembers, isclass
from pathlib import Path

import yaml

from simmate.configuration import settings
from simmate.engine import Workflow
from simmate.utilities import get_app_submodule, get_directory, make_archive


def get_all_workflows(
    apps_to_search: list[str] = settings.apps,
    as_dict: bool = False,
    exclude_subflows: bool = False,
) -> list[Workflow]:
    """
    Goes through a list of apps and grabs all workflow objects available.
    By default, this will grab all installed 'settings.apps'
    """
    app_workflows = []
    for app_name in apps_to_search:
        # check if there is a workflow module for this app and load it if so
        workflow_path = get_app_submodule(app_name, "workflows")
        if not workflow_path:
            continue  # skip to the next app
        app_workflow_module = importlib.import_module(workflow_path)

        # iterate through each available object in the workflows file and find
        # which ones are workflow objects.

        # If an __all__ value is set, then this will take priority when grabbing
        # workflows from the module
        if hasattr(app_workflow_module, "__all__"):
            for workflow_name in app_workflow_module.__all__:
                workflow = getattr(app_workflow_module, workflow_name)
                if workflow not in app_workflows:
                    app_workflows.append(workflow)

        # otherwise we load ALL class objects from the module -- assuming the
        # user properly limited these to just Workflow objects.
        else:
            # a tuple is returned by getmembers so c[0] is the string name while
            # c[1] is the python class object.
            app_workflows += [
                c[1] for c in getmembers(app_workflow_module) if isclass(c[1])
            ]

    # I define subflows as any workflow that has prerequisites needed to run.
    # These workflows are not user friendly and often hidden from docs
    # and the web ui
    if exclude_subflows:
        app_workflows = [w for w in app_workflows if not w.has_prerequisite]

    return (
        app_workflows
        if not as_dict
        else {flow.name_full: flow for flow in app_workflows}
    )


def get_all_workflow_names(
    apps_to_search: list[str] = settings.apps,
    exclude_subflows: bool = False,
) -> list[str]:
    """
    Returns a list of all the workflows of all types.
    """
    flow_names = [
        flow.name_full
        for flow in get_all_workflows(
            apps_to_search=apps_to_search,
            exclude_subflows=exclude_subflows,
        )
    ]
    flow_names.sort()
    return flow_names


def get_all_workflow_types(exclude_subflows: bool = False) -> list[str]:
    """
    Grabs all workflow types, which is also all "Project Names" and all of the
    submodules listed in the `simmate.workflows` module.

    Gives names in the format with all lowercase and hypens, such as "relaxation"
    or "static-energy"
    """
    workflow_types = []

    for flow in get_all_workflows(exclude_subflows=exclude_subflows):
        if flow.name_type not in workflow_types:
            workflow_types.append(flow.name_type)

    workflow_types.sort()

    return workflow_types


def get_apps_by_type(
    flow_type: str,
    precheck_type_exists: bool = True,
    exclude_subflows: bool = False,
) -> list[str]:
    """
    Returns a list of all the available apps for a given workflow type.
    """

    # Make sure the type is supported
    if precheck_type_exists:
        workflow_types = get_all_workflow_types(exclude_subflows=exclude_subflows)
        if flow_type not in workflow_types:
            raise TypeError(
                f"{flow_type} is not allowed. Please use a workflow type from "
                f"this list: {workflow_types}"
            )

    app_names = []
    for flow in get_all_workflows(exclude_subflows=exclude_subflows):
        if flow.name_type == flow_type and flow.name_app not in app_names:
            app_names.append(flow.name_app)

    app_names.sort()
    return app_names


def get_workflow_names_by_type(
    flow_type: str,
    app_name: str = None,
    full_name: bool = True,
    precheck_type_exists: bool = True,
    remove_no_database_flows: bool = False,
    exclude_subflows: bool = False,
) -> list[str]:
    """
    Returns a list of all the workflows of a given type. Optionally, the
    workflows can also be filtered by app name.
    """

    # Make sure the type is supported
    if precheck_type_exists:
        workflow_types = get_all_workflow_types(exclude_subflows=exclude_subflows)
        if flow_type not in workflow_types:
            raise TypeError(
                f"{flow_type} is not allowed. Please use a workflow type from "
                f"this list: {workflow_types}"
            )

    workflow_names = []

    for flow in get_all_workflows(exclude_subflows=exclude_subflows):
        if flow.name_type != flow_type:
            continue
        if app_name and flow.name_app != app_name:
            continue  # Skip those that don't match

        if remove_no_database_flows and not flow.use_database:
            continue

        if full_name:
            workflow_name = flow.name_full
        else:
            workflow_name = flow.name_preset

        workflow_names.append(workflow_name)

    workflow_names.sort()
    return workflow_names


def get_workflow(workflow_name: str) -> Workflow:
    """
    This is a utility for that grabs a workflow from the simmate workflows.

    This is typically used when we need to dynamically import a workflow, such
    as when calling a workflow through the command-line. We recommend avoiding
    this function when possible, and instead directly import your workflow.

    For example...
    ``` python
    from simmate.workflows.utilities import get_workflow

    matproj_workflow = get_workflow("static-energy.vasp.matproj")
    ```

    ...does the same exact thing as...
    ``` python
    from simmate.workflows.static_energy import StaticEnergy__Vasp__Matproj
    ```

    Note the naming of workflows therefore follows the format:
    ``` python

    # an example import of some workflow
    from simmate.workflows.example_module import example_workflow_class

    ```

    #### Parameters

    - `workflow_name`:
        Name of the workflow to load. Examples include "relaxation.vasp.matproj",
        "static-energy.vasp.quality01", and "diffusion.vasp.all-paths"
    """

    # First check if the user is providing a custom workflow, where the path is
    # given as "path/to/my/script:my_workflow_obj". If so, we need to load that
    # file and grab the workflow like it's an object from a python module.
    if ":" in workflow_name:
        # Note, windows paths have multiple ":" so we need to use rsplit.
        script_name, workflow_obj_name = workflow_name.rsplit(":", 1)
        script_name = Path(script_name)
        module_name = script_name.stem

        # No idea how this works, but its from the official docs... This just
        # loads our python file as if was a module
        # https://stackoverflow.com/questions/19009932/
        # https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
        spec = importlib.util.spec_from_file_location(module_name, script_name)
        module = importlib.util.module_from_spec(spec)
        sys.modules[module_name] = module
        spec.loader.exec_module(module)

        # now grab the workflow from the module!
        workflow = getattr(module, workflow_obj_name)

        return workflow

    # otherwise the app should be registered and available in the settings.apps
    workflow_dict = get_all_workflows(as_dict=True)
    workflow = workflow_dict.get(workflow_name, None)

    # make sure we have a proper workflow name provided and were able to load
    # it successfully
    if not workflow:
        raise Exception(
            "The workflow you provided isn't known. Make sure you don't have any "
            "typos! If you want a list of all available workflows, use the command "
            "`simmate workflows list-all`. You can also interactively explore "
            "workflows with `simmate workflows explore`"
        )

    return workflow


def reverse_workflow_name(workflow_name: str) -> str:
    """
    Converts the workflow name back into the pythonic class name. For example,
    this converts "static-energy.vasp.mit" to "StaticEnergy__Vasp__Mit"
    """

    # parse the workflow name. (e.g. static-energy.vasp.mit --> static_energy + vasp + mit)
    project_name, app_name, preset_name = workflow_name.replace("-", "_").split(".")

    # Combine the names into the full class name
    # (e.g. static_energy + vasp + mit --> StaticEnergy__Vasp__Mit)
    workflow_class_name = "__".join(
        [n.title().replace("_", "") for n in [project_name, app_name, preset_name]]
    )

    return workflow_class_name


def get_unique_parameters() -> list[str]:
    """
    Returns a list of all unique parameter names accross all workflows.

    This utility is really just to help developers ensure they are covering
    all cases when implementing new features, so this function isn't actually
    called elsewhere.
    """

    workflows = get_all_workflows()
    unique_parameters = []
    for workflow in workflows:
        for parameter in workflow.parameter_names:
            if parameter not in unique_parameters and parameter not in [
                "kwargs",
                "cls",
            ]:
                unique_parameters.append(parameter)

    # for consistency, we sort these alphabetically
    unique_parameters.sort()

    return unique_parameters


def load_results_from_directories(base_directory: Path | str = "."):
    """
    Goes through a given directory and finds all "simmate-task-" folders and zip
    archives present. The simmate_metadata.yaml file is used in each of these
    to load results into the database. All folders will be converted to archives
    once they've been loaded.

    #### Parameters

    - `base_directory`:
        The main directory that will contain folders to archive. Defaults to the
        working directory.
    """
    # load the full path to the desired directory
    directory = get_directory(base_directory)

    # grab all files/folders in this directory and then limit this list to those
    # that are...
    #   1. folders
    #   2. start with "simmate-task-"
    #   3. haven't been modified for at least time_cutoff
    foldernames = [
        directory / f for f in directory.iterdir() if "simmate-task-" in f.name
    ]

    # Now go through this list and archive the folders that met the criteria
    # and load the data from each.
    for foldername in foldernames:
        # Print message for monitoring progress.
        logging.info(f"Loading data from {foldername}")

        # There may be many folders that contain failed or incomplete data.
        # We don't want those to prevent others from being loaded so we put
        # everything in a try/except.
        try:
            # If we have a zip file, we need to unpack it before we can read results
            if not foldername.is_dir():
                shutil.unpack_archive(
                    filename=foldername,
                    extract_dir=directory,
                )
                # remove the ".zip" ending for our folder
                foldername = foldername.removesuffix(".zip")

            # Grab the metadata file which tells us key information
            filename = foldername / "simmate_metadata_01.yaml"
            with filename.open() as file:
                metadata = yaml.full_load(file)

            # see which workflow was used -- which also tells us the database table
            workflow_name = metadata["workflow_name"]
            workflow = get_workflow(workflow_name)

            # now load the data
            results_db = workflow.database_table.from_directory(foldername)

            # use the metadata to update the other fields
            results_db.source = metadata["source"]
            results_db.run_id = metadata["run_id"]

            # note the directory might have been moved from when this was originally
            # ran vs where it is now. Therefore, we update the folder location here.
            results_db.directory = foldername

            # Now save the results and convert the folder to an archive
            results_db.save()
            make_archive(foldername)

            logging.info("Successfully loaded into database.")

        except:
            logging.warning("Failed to load into database")



# -*- coding: utf-8 -*-

from simmate.apps.vasp.error_handlers import (  # Frozen,
    Brions,
    Brmix,
    Edddav,
    Eddrmm,
    Edwav,
    ElfKpar,
    IncorrectShift,
    IncorrectSmearing,
    InsufficientBands,
    LargeSigma,
    LongVector,
    MeshSymmetry,
    NonConverging,
    PointGroup,
    PositiveEnergy,
    Posmap,
    Potim,
    Pricel,
    Pssyevx,
    RealOptlay,
    Rhosyg,
    RotationMatrix,
    RotationNonIntMatrix,
    SubspaceMatrix,
    SymprecNoise,
    Tetirr,
    TetrahedronMesh,
    TripleProduct,
    Unconverged,
    Walltime,
    Zbrent,
    Zheev,
    Zpotrf,
)
from simmate.apps.vasp.inputs.potcar_mappings import PBE_POTCAR_MAPPINGS
from simmate.apps.vasp.workflows.base import VaspWorkflow


class Relaxation__Vasp__WarrenLabPbe(VaspWorkflow):
    """
    Runs a VASP geometry optimization using preferred settings from the Warren lab.

    """

    description_doc_short = "Warren Lab's presets for PBE optimization"

    functional = "PBE"
    potcar_mappings = PBE_POTCAR_MAPPINGS
    # For now I'm keeping the matproj default element mappings, but I will likely
    # update these for future calculations.

    _incar = dict(
        ALGO="Fast",
        EDIFF__per_atom=5e-05,  # This is slightly higher quality than the settings in Simmate (5e-05)
        ENCUT=520,  # Should set dynamically in future
        IBRION=2,
        ISIF=3,
        ISMEAR=0,  # Gaussian smearing
        ISPIN=2,
        KSPACING=0.25,  # This is probably on the edge of good quality
        LASPH=True,
        LORBIT=11,
        LREAL="Auto",
        LWAVE=False,
        NELM=150,
        NSW=99,
        PREC="Accurate",
        SIGMA=0.05,
        IVDW=12,  # adds van der waals correction
        MAGMOM__smart_magmom={
            "default": 0.6,  # note this is different from the VASP default
            "Ce": 5,
            "Ce3+": 1,
            "Co": 0.6,
            "Co3+": 0.6,
            "Co4+": 1,
            "Cr": 5,
            "Dy3+": 5,
            "Er3+": 3,
            "Eu": 10,
            "Eu2+": 7,
            "Eu3+": 6,
            "Fe": 5,
            "Gd3+": 7,
            "Ho3+": 4,
            "La3+": 0.6,
            "Lu3+": 0.6,
            "Mn": 5,
            "Mn3+": 4,
            "Mn4+": 3,
            "Mo": 5,
            "Nd3+": 3,
            "Ni": 5,
            "Pm3+": 4,
            "Pr3+": 2,
            "Sm3+": 5,
            "Tb3+": 6,
            "Tm3+": 2,
            "V": 5,
            "W": 5,
            "Yb3+": 1,
        },
        LMAXMIX__smart_lmaxmix=True,
        multiple_keywords__smart_ldau=dict(
            LDAU__auto=True,
            LDAUTYPE=2,
            LDAUPRINT=1,
            LDAUJ={},  # pymatgen sets these, but they're all default values anyways
            LDAUL={
                "F": {
                    "Co": 2,
                    "Cr": 2,
                    "Fe": 2,
                    "Mn": 2,
                    "Mo": 2,
                    "Ni": 2,
                    "V": 2,
                    "W": 2,
                },
                "O": {
                    "Co": 2,
                    "Cr": 2,
                    "Fe": 2,
                    "Mn": 2,
                    "Mo": 2,
                    "Ni": 2,
                    "V": 2,
                    "W": 2,
                },
            },
            LDAUU={
                "F": {
                    "Co": 3.32,
                    "Cr": 3.7,
                    "Fe": 5.3,
                    "Mn": 3.9,
                    "Mo": 4.38,
                    "Ni": 6.2,
                    "V": 3.25,
                    "W": 6.2,
                },
                "O": {
                    "Co": 3.32,
                    "Cr": 3.7,
                    "Fe": 5.3,
                    "Mn": 3.9,
                    "Mo": 4.38,
                    "Ni": 6.2,
                    "V": 3.25,
                    "W": 6.2,
                },
            },
        ),
    )

    error_handlers = [
        TetrahedronMesh(),
        RotationMatrix(),
        Brmix(),
        Zpotrf(),
        SubspaceMatrix(),
        IncorrectShift(),
        RealOptlay(),
        Tetirr(),
        RotationNonIntMatrix(),
        LongVector(),
        TripleProduct(),
        Pricel(),
        Brions(),
        Zbrent(),
        InsufficientBands(),
        Pssyevx(),
        Eddrmm(),
        Edddav(),
        Edwav(),
        Zheev(),
        ElfKpar(),
        Rhosyg(),
        Posmap(),
        PointGroup(),
        SymprecNoise(),
        IncorrectSmearing(),
        MeshSymmetry(),
        Unconverged(),
        NonConverging(),
        Potim(),
        PositiveEnergy(),
        #        Frozen(),
        LargeSigma(),
        Walltime(),
    ]



# -*- coding: utf-8 -*-

from simmate.database.base_data_types import Structure, Thermodynamics, table_column


class MatprojStructure(Structure, Thermodynamics):
    """
    Crystal structures from the [Materials Project](https://materialsproject.org/)
    database.

    Currently, this table only stores strucure and thermodynamic information,
    but the Materials Project has much more data available via their
    [REST API](https://github.com/materialsproject/api) and website.
    """

    class Meta:
        app_label = "data_explorer"

    archive_fields = [
        "energy_uncorrected",
        "band_gap",
        "is_gap_direct",
        "is_magnetic",
        "total_magnetization",
        "is_theoretical",
    ]
    api_filters = dict(
        energy_uncorrected=["range"],
        band_gap=["range"],
        is_gap_direct=["exact"],
        is_magnetic=["exact"],
        total_magnetization=["range"],
        is_theoretical=["exact"],
    )
    source = "Materials Project"
    source_long = "The Materials Project at Berkeley National Labs"
    homepage = "https://materialsproject.org/"
    source_doi = "https://doi.org/10.1063/1.4812323"
    remote_archive_link = "https://archives.simmate.org/MatprojStructure-2023-07-07.zip"

    id = table_column.CharField(max_length=25, primary_key=True)
    """
    The id used to represent the structure (ex: "mp-12345")
    """

    energy_uncorrected = table_column.FloatField(blank=True, null=True)
    """
    The reported energy of the system BEFORE Materials Project applies
    their composition-based corrections.
    """

    band_gap = table_column.FloatField(blank=True, null=True)
    """
    The band gap energy in eV.
    """

    is_gap_direct = table_column.BooleanField(blank=True, null=True)
    """
    Whether the band gap is direct or indirect.
    """

    is_magnetic = table_column.BooleanField(blank=True, null=True)
    """
    Whether the material is magnetic
    """

    total_magnetization = table_column.FloatField(blank=True, null=True)
    """
    The total magnetization of the material
    """

    is_theoretical = table_column.BooleanField(blank=True, null=True)
    """
    Whether the material is from a theoretical structure. False indicates
    that it is experimentally known.
    """

    updated_at = table_column.DateTimeField(blank=True, null=True)
    """
    Timestamp of when this row was was lasted changed / updated by the 
    Materials Project
    """

    @property
    def external_link(self) -> str:
        """
        URL to this structure in the Materials Project website.
        """
        # All Materials Project structures have their data mapped to a URL in
        # the same way. For example...
        #   https://materialsproject.org/materials/mp-12345/
        return f"https://materialsproject.org/materials/{self.id}/"

    # -------------------------------------------------------------------------

    @classmethod
    def _load_all_structures(
        cls,
        api_key: str,
        update_stabilities: bool = False,
    ):
        """
        Only use this function if you are part of the Simmate dev team!
        Users should instead access data via the load_remote_archive method.

        This method is for pulling Materials Project data into the Simmate database.


        PyMatGen offers an easy way to do this in python -- the MPRester class.
        All you need is [an API key from their site](https://materialsproject.org/open)
        and pymatgen installed.

        #### Parameters

        - `api_key`:
            Your Materials Project API key.
        - `criteria`:
            Filtering criteria for which structures to load. The default is all
            existing structures (137,885 as of 2022-01-16), which will take rouhghly
            15 min to complete (not including stabilities).
        - `update_stabilities`:
            Whether to run update_all_stabilities on the database table. Note this
            will add over an hour to this process. Default is True.
        """

        from django.utils.timezone import make_aware
        from rich.progress import track

        try:
            from mp_api.client import MPRester
        except:
            raise Exception(
                "To use this method, MP-API is required. Please install it "
                "with `pip install mp-api"
            )

        # Connect to their database with personal API key
        with MPRester(api_key) as mpr:
            # For the filtered structures, this lists off which properties to grab.
            # All possible properties can be listed with:
            #   mpr.summary.available_fields
            fields_to_load = [
                # "builder_meta",
                # "nsites",
                # "elements",
                # "nelements",
                # "composition",
                # "composition_reduced",
                # "formula_pretty",
                # "formula_anonymous",
                # "chemsys",
                # "volume",
                # "density",
                # "density_atomic",
                # "symmetry",
                # "property_name",
                "material_id",
                # "deprecated",
                # "deprecation_reasons",
                "last_updated",
                # "origins",
                # "warnings",
                "structure",
                # "task_ids",
                "uncorrected_energy_per_atom",
                "energy_per_atom",
                # "formation_energy_per_atom",
                # "energy_above_hull",
                # "is_stable",
                # "equilibrium_reaction_energy_per_atom",
                # "decomposes_to",
                # "xas",
                # "grain_boundaries",
                "band_gap",
                # "cbm",
                # "vbm",
                # "efermi",
                "is_gap_direct",
                # "is_metal",
                # "es_source_calc_id",
                # "bandstructure",
                # "dos",
                # "dos_energy_up",
                # "dos_energy_down",
                "is_magnetic",
                # "ordering",
                "total_magnetization",
                # "total_magnetization_normalized_vol",
                # "total_magnetization_normalized_formula_units",
                # "num_magnetic_sites",
                # "num_unique_magnetic_sites",
                # "types_of_magnetic_species",
                # "k_voigt",
                # "k_reuss",
                # "k_vrh",
                # "g_voigt",
                # "g_reuss",
                # "g_vrh",
                # "universal_anisotropy",
                # "homogeneous_poisson",
                # "e_total",
                # "e_ionic",
                # "e_electronic",
                # "n",
                # "e_ij_max",
                # "weighted_surface_energy_EV_PER_ANG2",
                # "weighted_surface_energy",
                # "weighted_work_function",
                # "surface_anisotropy",
                # "shape_factor",
                # "has_reconstructed",
                # "possible_species",
                # "has_props",
                "theoretical",
            ]

            # now make the query and grab everything from the Materials Project!
            # the output dictionary is given back within a list, where each entry is
            # a specific structure (so a single mp-id)
            # Note: this is a very large query, so make sure your computer has enough
            # memory (RAM >10GB) and a stable internet connection.
            data = mpr.summary.search(
                all_fields=False,
                fields=fields_to_load,
                deprecated=False,
                # !!! DEV NOTE: you can uncomment these lines for quick testing
                # num_chunks=3,
                chunk_size=1000,
            )

            # BUG: The search above is super unstable, so instead, I grab all mp-id
            # in one search, then make individual queries for the data of each
            # after that.
            # This takes about 30 minutes.
            # mp_ids = mpr.summary.search(
            #     all_fields=False,
            #     fields=["material_id"],
            #     deprecated=False,
            # )
            # data = []
            # chunk_ids = []
            # for entry in track(mp_ids):
            #     chunk_ids.append(entry.material_id)
            #     if (
            #         len(chunk_ids) >= 1000
            #         or entry.material_id == mp_ids[-1].material_id
            #     ):
            #         result = mpr.summary.search(
            #             material_ids=[entry.material_id],
            #             all_fields=False,
            #             fields=fields_to_load,
            #         )
            #         data += result
            #         chunk_ids = []

        # Let's iterate through each structure and save it to the database
        # This also takes a while, so we use a progress bar
        failed_entries = []
        db_objects = []
        for entry in track(data):
            try:
                # convert the data to a Simmate database object
                structure_db = cls.from_toolkit(
                    id=str(entry.material_id),
                    structure=entry.structure,
                    energy=entry.energy_per_atom * entry.structure.num_sites,
                    energy_uncorrected=entry.uncorrected_energy_per_atom
                    * entry.structure.num_sites,
                    updated_at=make_aware(entry.last_updated),
                    band_gap=entry.band_gap,
                    is_gap_direct=entry.is_gap_direct,
                    is_magnetic=entry.is_magnetic,
                    total_magnetization=entry.total_magnetization,
                    is_theoretical=entry.theoretical,
                )
                db_objects.append(structure_db)
            except:
                failed_entries.append(entry)

        # and save it to our database
        cls.objects.bulk_create(
            db_objects,
            batch_size=15000,
            ignore_conflicts=True,
        )

        # once all structures are saved, let's update the Thermodynamic columns
        if update_stabilities:
            cls.update_all_stabilities()

        return failed_entries



# -*- coding: utf-8 -*-

import logging
import warnings

from rich.progress import track

from simmate.database.base_data_types import DatabaseTable, table_column
from simmate.toolkit import Structure as ToolkitStructure
from simmate.utilities import get_chemical_subsystems
from simmate.visualization.plotting import PlotlyFigure

# BUG: This prints a tqdm error so we silence it here.
with warnings.catch_warnings(record=True):
    from pymatgen.analysis.phase_diagram import PDEntry, PDPlotter, PhaseDiagram


class Thermodynamics(DatabaseTable):
    class Meta:
        abstract = True

    archive_fields = ["energy"]

    api_filters = dict(
        energy=["range"],
        energy_per_atom=["range"],
        energy_above_hull=["range"],
        is_stable=["exact"],
        formation_energy=["range"],
        formation_energy_per_atom=["range"],
    )

    energy = table_column.FloatField(blank=True, null=True)
    """
    The calculated total energy. Units are in eV.
    """

    energy_per_atom = table_column.FloatField(blank=True, null=True)
    """
    The `energy` divided by `nsites`. Units are in eV.
    """

    # TODO: These columns below should be updated on a schedule. I'd run
    # a prefect flow every night that makes sure all values are up to date.

    energy_above_hull = table_column.FloatField(blank=True, null=True)
    """
    The hull energy (aka "stability") of this structure compared to all other
    structures in this database table. Units are in eV.
    """

    is_stable = table_column.BooleanField(blank=True, null=True)
    """
    Whether `energy_above_hull` is 0 -- if so, this is considered stable.
    """

    decomposes_to = table_column.JSONField(blank=True, null=True)
    """
    If `energy_above_hull` is above 0, these are the compositions that the 
    structure will decompose to (e.g. ["Y2C", "C", "YF3"])
    """

    formation_energy = table_column.FloatField(blank=True, null=True)
    """
    The formation energy of the structure relative to all other
    structures in this database table. Units are in eV.
    """

    formation_energy_per_atom = table_column.FloatField(blank=True, null=True)
    """
    The `formation_energy` divided by `nsites`.
    """

    # Other fields to consider
    # equilibrium_reaction_energy_per_atom
    # energy_uncertainy_per_atom
    # energy_per_atom_uncorrected
    # decompose_amount (to each decomp type)

    """ Properties """
    # These are some extra fields to consider adding
    # energy_type = ... (DFT, machine-learned, LJ, etc.)
    # calc_types = (list) https://github.com/materialsproject/emmet/blob/main/emmet-core/emmet/core/vasp/calc_types/utils.py
    # psudeopotential
    #     functional(PBE)
    #     label(Y_sv)
    #     pot_type(PAW)
    # type(GGA,GGAU,HF)
    # is_hubbard

    @classmethod
    def _from_toolkit(
        cls,
        structure: ToolkitStructure | str = None,
        energy: float = None,
        as_dict: bool = False,
    ):
        if isinstance(structure, str):
            structure = ToolkitStructure.from_database_string(structure)

        # Given energy, this function builds the rest of the required fields
        # for this class as an object (or as a dictionary).
        data = dict(energy=energy) if energy else {}

        # if a structure is present, we can update that information as well.
        if structure and energy:
            epa_data = {"energy_per_atom": energy / structure.num_sites}
            data.update(epa_data)

        # OPTIMIZE: I try calculating these when each structure is added, but
        # this would be too slow. Instead, I have the user call the
        # update_all_stabilites method on a cycle.
        # energy_above_hull=None,
        # is_stable=None,
        # decomposes_to=None,

        # If as_dict is false, we build this into an Object. Otherwise, just
        # return the dictionary
        return data if as_dict else cls(**data)

    @classmethod
    def update_chemical_system_stabilities(
        cls,
        chemical_system: str,
        workflow_name: str = None,
    ):
        phase_diagram, entries, entries_pmg = cls.get_phase_diagram(
            chemical_system,
            return_entries=True,
            workflow_name=workflow_name,
        )

        # now go through the entries and update stability values
        for entry, entry_pmg in zip(entries, entries_pmg):
            decomp, hull_energy = phase_diagram.get_decomp_and_e_above_hull(entry_pmg)

            entry.energy_above_hull = hull_energy

            entry.is_stable = True if hull_energy == 0 else False

            # OPTIMIZE: I would like this to point to another entry specifically
            # but this will take more work.
            entry.decomposes_to = (
                [d.composition.formula for d in decomp] if hull_energy != 0 else []
            )

            entry.formation_energy = phase_diagram.get_form_energy(entry_pmg)
            entry.formation_energy_per_atom = phase_diagram.get_form_energy_per_atom(
                entry_pmg
            )

        # Now that we updated our objects, we want to collectively update them
        cls.objects.bulk_update(
            objs=entries,
            fields=[
                "energy_above_hull",
                "is_stable",
                "decomposes_to",
                "formation_energy",
                "formation_energy_per_atom",
            ],
            # updating extremely large systems (>3k structures) can cause this
            # to time-out and crash. We therefore update in batches of 500
            batch_size=500,
        )

    @classmethod
    def update_all_stabilities(cls, workflow_name: str = None):
        # grab all unique chemical systems
        chemical_systems = cls.objects.values_list(
            "chemical_system", flat=True
        ).distinct()

        # Now  go through each and run the analysis.
        # OPTIMIZE: Some systems will be analyzed multiple times. For example,
        # C would be repeatedly updated through C, C-O, Y-C-F, etc.
        for chemical_system in track(chemical_systems):
            try:
                cls.update_chemical_system_stabilities(
                    chemical_system,
                    workflow_name,
                )
            except ValueError as exception:
                logging.warning(f"Failed for {chemical_system} with error: {exception}")

    @classmethod
    def get_phase_diagram(
        cls,
        chemical_system: str,
        workflow_name: str = None,
        return_entries: bool = False,
    ) -> PhaseDiagram:
        if workflow_name is None and hasattr(cls, "workflow_name"):
            raise Exception(
                "This table contains results from multiple workflows, so you must "
                "provide a workflow_name as an input to indicate which entries "
                "should be loaded/updated."
            )

        # if we have a multi-element system, we need to include subsystems as
        # well. ex: Na --> Na, Cl, Na-Cl
        subsystems = get_chemical_subsystems(chemical_system)

        # grab all entries for this chemical system
        entries = cls.objects.filter(
            # workflow_name="relaxation.vasp.staged",
            chemical_system__in=subsystems,
            energy__isnull=False,  # only completed calculations
        )
        # add an extra filter if provided
        if workflow_name:
            entries = entries.filter(workflow_name=workflow_name)

        # now make the queryy
        entries = entries.only("id", "energy", "formula_full").all()

        # convert to pymatgen PDEntries and build into PhaseDiagram object
        entries_pmg = []
        for entry in entries:
            pde = PDEntry(
                composition=entry.formula_full,
                energy=entry.energy,
                # name=entry.id,  see bug below
            )

            # BUG: pymatgen grabs entry_id, when it should really be grabbing name.
            # https://github.com/materialsproject/pymatgen/blob/de17dd84ba90dbf7a8ed709a33d894a4edb82d02/pymatgen/analysis/phase_diagram.py#L2926
            pde.entry_id = f"id={entry.id}"
            entries_pmg.append(pde)

        phase_diagram = PhaseDiagram(entries_pmg)

        return (
            phase_diagram
            if not return_entries
            else (
                phase_diagram,
                entries,
                entries_pmg,
            )
        )


class HullDiagram(PlotlyFigure):
    method_type = "classmethod"

    def get_plot(
        table,  # Thermodynamics + Structure table
        chemical_system: str,
        workflow_name: str = None,
        show_unstable_up_to: float = float("inf"),
    ):
        phase_diagram = table.get_phase_diagram(chemical_system, workflow_name)

        # alternatively use backend="matplotlib"
        plotter = PDPlotter(phase_diagram, show_unstable=show_unstable_up_to)
        # Only shows structures up to X eV above hull

        plot = plotter.get_plot(label_unstable=False)

        return plot


# register all plotting methods to the database table
for _plot in [HullDiagram]:
    _plot.register_to_class(Thermodynamics)



# -*- coding: utf-8 -*-

from ase import Atoms
from ase.ga.startgenerator import StartGenerator
from ase.ga.utilities import CellBounds, closest_distances_generator
from pymatgen.io.ase import AseAtomsAdaptor

from simmate.toolkit import Composition, Structure
from simmate.toolkit.creators.structure.base import StructureCreator


class AseStructure(StructureCreator):
    """
    Creates structures using ASE's start-generator

    see source: https://gitlab.com/ase/ase/-/blob/master/ase/ga/startgenerator.py
    see tutorials: https://wiki.fysik.dtu.dk/ase/ase/ga/ga.html#module-ase.ga
    """

    def __init__(
        self,
        composition: Composition,
        ratio_of_covalent_radii: int = 0.5,
    ):
        # I assume 3D structure for now. I could do things like pbc=[1,1,0]
        # for 2D in the future though
        slab = Atoms(pbc=True)

        # the blocks input is just a list of building blocks and how many to
        # include of each for example, Ca2N would have [('Ca', 2), ('N', 1)].
        # let's convert the pymatgen composition object to this format - other
        # formats are allowed too
        blocks = [
            (element.symbol, int(composition[element])) for element in composition
        ]

        # the closest_distances_generator is exactly the same as an
        # element-dependent distance matrix. ASE puts this in dictionary form
        # the function requires a list of element integers
        element_ints = [element.number for element in composition]
        # the default of the ratio of covalent radii (0.5) is based on the
        # ASE tutorial of this function
        element_distance_matrix = closest_distances_generator(
            element_ints, ratio_of_covalent_radii
        )

        # boundry limits on the lattice
        # I only do angle limits for now because I try to set the volume below
        cellbounds = CellBounds(
            bounds={
                "phi": [60, 120],
                "chi": [60, 120],
                "psi": [60, 120],
            }
        )
        #'a': [3, 50], 'b': [3, 50], 'c': [3, 50]})

        # predict volume
        volume_guess = composition.volume_estimate()

        self.ase = StartGenerator(
            slab=slab,
            blocks=blocks,
            blmin=element_distance_matrix,
            # !!! I understand this as number of unique vectors. Is that right?
            number_of_variable_cell_vectors=3,
            # box_to_place_in=None, # use the entire unitcell
            box_volume=volume_guess,  # target this volume when making the lattice
            cellbounds=cellbounds,
            # test_dist_to_slab=True, # ensure distances arent too close
            # test_too_far=True, # ensure no atom is isolated
            # rng=np.random
        )

    def create_structure(self) -> Structure:
        # now make the new structure using ase,ga.startgenerator
        # NOTE: all of these options are set in the init except for spacegroup
        structure_ase = self.ase.get_new_candidate(maxiter=1000)

        # if creation fails, None is returned
        if not structure_ase:
            return False

        # convert the ase Atoms object to pymatgen Structure
        structure = AseAtomsAdaptor.get_structure(structure_ase)

        return structure



# -*- coding: utf-8 -*-

import datetime
import importlib
import logging
import time
from traceback import format_exc

from django.core.mail import EmailMessage
from rich import print
from schedule import Scheduler

from simmate.configuration import settings
from simmate.utilities import get_app_submodule

# This string is just something fancy to display in the console when the process
# starts up.
# This uses "Small Slant" from https://patorjk.com/software/taag/
HEADER_ART = r"""
=========================================================================
   _____                  __        ____    __          __     __       
  / __(_)_ _  __ _  ___ _/ /____   / __/___/ /  ___ ___/ /_ __/ /__ ____
 _\ \/ /  ' \/  ' \/ _ `/ __/ -_) _\ \/ __/ _ \/ -_) _  / // / / -_) __/
/___/_/_/_/_/_/_/_/\_,_/\__/\__/ /___/\__/_//_/\__/\_,_/\_,_/_/\__/_/   
                                                                        
=========================================================================
"""


class SimmateScheduler(Scheduler):
    """
    Starts the main process for periodic tasks in each app's "schedules" module.

    NOTE: This is a "basic" alternative to scheduler systems such as Prefect.
    Here, only 1 task is ran at a time. So if you have a >1 hr task, this will
    block tasks schedules to run every minute until the longer task finishes.
    Furthermore, if you schedule a task to run at an exaction date/time, this
    scheduler may miss the start time due to other running tasks in front of it.
    Lastly, we "sleep" the scheduler every second, so scheduling tasks to run
    every <1s will not work as intended. The 1s sleep also means start times
    will have an error of up to 1s -- even when there's only one scheduled task.
    """

    def __init__(self, reschedule_on_failure=True):
        """
        If reschedule_on_failure is True, jobs will be rescheduled for their
        next run as if they had completed successfully. If False, they'll run
        on the next run_pending() tick.
        """
        self.reschedule_on_failure = reschedule_on_failure
        super().__init__()

    @classmethod
    def start(cls, sleep_step: float = 1):
        """
        Starts the main process for periodic tasks in each app's "schedules" module.

        NOTE: This is a "basic" alternative to scheduler systems such as Prefect.
        Here, only 1 task is ran at a time. So if you have a >1 hr task, this will
        block tasks schedules to run every minute until the longer task finishes.
        Furthermore, if you schedule a task to run at an exaction date/time, this
        scheduler may miss the start time due to other running tasks in front of it.
        Lastly, we "sleep" the scheduler every second, so scheduling tasks to run
        every <1s will not work as intended. The 1s sleep also means start times
        will have an error of up to 1s -- even when there's only one scheduled task.
        """

        # TODO: consider parallel runs using threads, dask, or workers...
        # https://schedule.readthedocs.io/en/stable/parallel-execution.html

        # TODO: for error handling, read more at...
        # https://schedule.readthedocs.io/en/stable/exception-handling.html

        # print the header in the console to let the user know the worker started
        print("[bold cyan]" + HEADER_ART)

        # HACK-FIX:
        # Jobs will register to the default scheduler. We instead want to use
        # this custom class, so we patch the schedule module
        # https://github.com/dbader/schedule/blob/master/schedule/__init__.py#L801C1-L881C31
        import schedule

        schedule.default_scheduler = cls()

        # Now run the registration where the scheduler shortcuts will work
        cls._register_app_schedules()

        # And now run the infinite loop of schedules
        logging.info("Starting schedules...")
        while True:  # Run indefinitely
            schedule.run_pending()
            # save some CPU time by sleeping for an extra second
            time.sleep(sleep_step)

    @staticmethod
    def _register_app_schedules(apps_to_search: list[str] = settings.apps):
        """
        Goes through a list of apps and imports the "schedules" module in each.
        By default, this will grab all installed 'settings.apps'
        """
        logging.info("Searching for registrations...")
        for app_name in apps_to_search:
            # check if there is a schedule module for this app and load it if so
            schedule_path = get_app_submodule(app_name, "schedules")
            if not schedule_path:
                continue  # skip to the next app
            # We use the python 'schedule' package, so simply importing these
            # modules is enough to register them.
            importlib.import_module(schedule_path)
            logging.info(f"Registered schedules for '{schedule_path}'")
        logging.info("Completed registrations :sparkles:")

    def _run_job(self, job):
        # This is a modified run method that catches failed jobs and optionally
        # sends an email alert on failure events

        try:
            super()._run_job(job)
        except Exception:
            # log errors but still continue
            error_msg = format_exc()
            logging.critical(error_msg)
            job.last_run = datetime.datetime.now()
            job._schedule_next_run()

            # if emails are configured, send an alert of the failure
            email = EmailMessage(
                subject="[SIMMATE] Scheduled job failure",
                body=error_msg,
                to=[a[1] for a in settings.website.admins],  # get admin emails
            )
            email.send(fail_silently=True)



# -*- coding: utf-8 -*-

from simmate.apps.warren_lab.workflows.relaxation.pbe import (
    Relaxation__Vasp__WarrenLabPbe,
)

PBE_STATIC_SETTINGS = dict(
    IBRION=-1,  # (optional) locks everything between ionic steps
    NSW=0,  # this is the main static energy setting
    #            LAECHG=True, # currently only set in population analysis
    LCHARG=True,
    LORBIT=11,
    LVHAR=True,
    LWAVE=True,
    ALGO="Normal",  # was "Fast" before
)
# These settings are saved here to make it easier to add them to other static
# energy workflows


class StaticEnergy__Vasp__WarrenLabPbe(Relaxation__Vasp__WarrenLabPbe):
    """
    Performs a static energy calculation based on the settings for Warren Lab
    PBE.
    """

    _incar_updates = PBE_STATIC_SETTINGS



# -*- coding: utf-8 -*-

from simmate.apps.warren_lab.workflows.relaxation.hse import (
    Relaxation__Vasp__WarrenLabHse,
)


class Relaxation__Vasp__WarrenLabHsesol(Relaxation__Vasp__WarrenLabHse):
    """
    Runs a VASP relaxation calculation using Warren Lab HSEsol settings.
    """

    description_doc_short = "Warren Lab presets for HSEsol geometry optimization"

    _incar_updates = dict(
        IVDW=0,  # No d3 constants for HSEsol so we just turn of VDW corrections
        GGA="PS",  # Tells vasp to use PBEsol instead of PBE
    )



# -*- coding: utf-8 -*-

import itertools
import logging
import math
from pathlib import Path

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from numpy.typing import ArrayLike
from pymatgen.io.vasp import Poscar
from pyrho.pgrid import PGrid

from simmate.toolkit import Structure


class Grid:
    """
    A grid object. Contains information from a VASP CHGCAR or ELFCAR type file.
    This class is used heavily during

    Args:
        total (ArrayLike): A 3D array of values corresponding to the total
            charge density of a CHGCAR type file or the spin up portion of an
            ELFCAR file.
        diff: (ArrayLike): A 3D array of values corresponding to the
            magnetization density of a CHGCAR type file or the spin down portion
            of an ELFCAR file.
        structure (Structure): a pymatgen Structure type object.
        data_type (str): The type of data, either "elf" or "charge density"
    """

    def __init__(
        self,
        total: ArrayLike,
        diff: ArrayLike | None,
        structure: Structure,
        data_type: str,
    ):
        if diff is not None and diff.shape != total.shape:
            raise ValueError("total and diff arrays must have the same shape.")
        self.total = total
        self.diff = diff
        self.structure = structure
        if data_type not in ["elf", "charge density", "unknown"]:
            raise ValueError("data type must be 'elf', 'charge density', or 'unknown'")
        self.data_type = data_type

    @property
    def grid_shape(self):
        """Data grid shape

        Returns:
            The number of voxels along each unit cell vector
        """
        return np.array(self.total.shape)

    @property
    def a(self):
        """The cartesian coordinates for the lattice vector "a"

        Returns:
            The cartesian coordinates for the lattice vector "a"
        """
        return self.structure.lattice.matrix[0]

    @property
    def b(self):
        """The cartesian coordinates for the lattice vector "b"

        Returns:
            The cartesian coordinates for the lattice vector "b"
        """
        return self.structure.lattice.matrix[1]

    @property
    def c(self):
        """The cartesian coordinates for the lattice vector "c"

        Returns:
            The cartesian coordinates for the lattice vector "c"
        """
        return self.structure.lattice.matrix[2]

    @property
    def frac_coords(self):
        """The fractional coordinates of every atom in the structure

        Returns:
            Array of coordinates for each atom.
        """
        return self.structure.frac_coords

    @property
    def voxel_volume(self):
        """The volume of each voxel in the grid

        Returns:
            The volume of a voxel
        """
        volume = self.structure.volume
        voxel_num = np.prod(self.grid_shape)
        return volume / voxel_num

    @property
    def voxel_num(self):
        return self.grid_shape.prod()

    @property
    def max_voxel_dist(self):
        """
        Finds the maximum distance a voxel can be from a dividing plane that still
        allows for the possibility that the voxel is intercepted by the plane.

        Returns:
            The maximum distance a voxel can be from a dividing plane and still
            be intercepted by the plane.
        """
        # We need to find the coordinates that make up a single voxel. This
        # is just the cartesian coordinates of the unit cell divided by
        # its grid size
        end = [0, 0, 0]
        vox_a = [x / self.grid_shape[0] for x in self.a]
        vox_b = [x / self.grid_shape[1] for x in self.b]
        vox_c = [x / self.grid_shape[2] for x in self.c]
        # We want the three other vertices on the other side of the voxel. These
        # can be found by adding the vectors in a cycle (e.g. a+b, b+c, c+a)
        vox_a1 = [x + x1 for x, x1 in zip(vox_a, vox_b)]
        vox_b1 = [x + x1 for x, x1 in zip(vox_b, vox_c)]
        vox_c1 = [x + x1 for x, x1 in zip(vox_c, vox_a)]
        # The final vertex can be found by adding the last unsummed vector to any
        # of these
        end1 = [x + x1 for x, x1 in zip(vox_a1, vox_c)]
        # The center of the voxel sits exactly between the two ends
        center = [(x + x1) / 2 for x, x1 in zip(end, end1)]
        # Shift each point here so that the origin is the center of the
        # voxel.
        voxel_vertices = []
        for vector in [center, end, vox_a, vox_b, vox_c, vox_a1, vox_b1, vox_c1, end]:
            new_vector = [(x - x1) for x, x1 in zip(vector, center)]
            voxel_vertices.append(new_vector)

        # Now we need to find the maximum distance from the center of the voxel
        # to one of its edges. This should be at one of the vertices.
        # We can't say for sure which one is the largest distance so we find all
        # of their distances and return the maximum
        max_distance = max([np.linalg.norm(vector) for vector in voxel_vertices])
        return max_distance

    @property
    def permutations(self):
        """
        The permutations for translating a voxel coordinate to nearby unit
        cells. This is necessary for the many voxels that will not be directly
        within an atoms partitioning.

        Returns:
            A list of voxel permutations unique to the grid dimensions.
        """
        a, b, c = self.grid_shape
        permutations = [
            (t, u, v)
            for t, u, v in itertools.product([-a, 0, a], [-b, 0, b], [-c, 0, c])
        ]
        # sort permutations. There may be a better way of sorting them. I
        # noticed that generally the correct site was found most commonly
        # for the original site and generally was found at permutations that
        # were either all negative/0 or positive/0
        permutations_sorted = []
        for item in permutations:
            if all(val <= 0 for val in item):
                permutations_sorted.append(item)
            elif all(val >= 0 for val in item):
                permutations_sorted.append(item)
        for item in permutations:
            if item not in permutations_sorted:
                permutations_sorted.append(item)
        permutations_sorted.insert(0, permutations_sorted.pop(7))
        return permutations_sorted

    @property
    def voxel_resolution(self):
        volume = self.structure.volume
        number_of_voxels = self.grid_shape.prod()
        return number_of_voxels / volume

    def get_grid_axes(self, padding: int = 0):
        """
        Gets the the possible indices for each dimension of a padded grid.
        e.g. if the original charge density grid is 20x20x20, and is padded
        with one extra layer on each side, this function will return three
        arrays with integers from 0 to 21.

        Args:
            padding (int): The amount the grid has been padded

        Returns:
            three arrays with lengths the same as the grids shape
        """
        grid = self.total
        a = np.linspace(
            0, grid.shape[0] + (padding - 1) * 2 + 1, grid.shape[0] + padding * 2
        )
        b = np.linspace(
            0, grid.shape[1] + (padding - 1) * 2 + 1, grid.shape[1] + padding * 2
        )
        c = np.linspace(
            0, grid.shape[2] + (padding - 1) * 2 + 1, grid.shape[2] + padding * 2
        )
        return a, b, c

    def copy(self):
        """
        Convenience method to get a copy of the grid.

        Returns:
            A copy of the Grid.
        """
        return self.__class__(
            self.total,
            self.diff,
            self.structure.copy(),
            self.data_type,
        )

    @classmethod
    def from_file(cls, grid_file: str | Path):
        """Create a grid instance using a CHGCAR or ELFCAR file

        Args:
            grid_file (string): The file the instance should be made from. Should
                be a VASP CHGCAR or ELFCAR type file.

        Returns:
            Grid from the specified file.
        """
        logging.info(f"Loading {grid_file} from file")
        # Create string to add structure to.
        structure_str = ""
        total_str = ""
        diff_str = ""

        # Read the content of the file. We will need the number of atoms and the
        # grid shape to read in the data.
        with open(grid_file) as f:
            content = f.readlines()

            # get the number of atoms in the structure
            num_atoms = sum([int(x) for x in content[6].split()])
            # get the grid shape of the data
            grid_shape_str = content[9 + num_atoms]
            grid_shape = [int(x) for x in grid_shape_str.split()]
            # get the first row of data. This will be used to determine the
            # format of the file.
            first_data_row = content[10 + num_atoms]

            # Get the structure as a string
            for line in content[0 : 8 + num_atoms]:
                structure_str += f"{line}"

            # The format that VASP outputs for CHGCARs and ELFCARs can have several
            # different forms. The data is effectively a long list of values going
            # in order of array indices.
            # Typically the CHGCAR will have 5 values per line while the
            # ELFCAR will have 10 if they are directly from a VASP calculation.
            # If they are outputted from pymatgen or this grid class they will both
            # have 5 columns. This will effect the number of lines of data there are.
            column_num = len(first_data_row.split())
            line_num = math.ceil(np.product(grid_shape) / column_num)
            skip_rows_total = 10 + num_atoms

            # Get the total grid as a string.
            for line in content[skip_rows_total : skip_rows_total + line_num]:
                total_str += f"{line}"

            # charge density file have a set of augmentation occupancies stored
            # after the charge density. We can look for a line containing this
            # to determine the data type. We also need to determine how many
            # lines this data takes up.
            aug_occ_lines_num = 0
            try:
                if "augmentation occupancies" in content[skip_rows_total + line_num]:
                    data_type = "charge density"
                    for line in content[skip_rows_total + line_num :]:
                        # If we find the start of a new set of data or the end of
                        # the file we end.
                        if grid_shape_str in line:
                            break
                        else:
                            aug_occ_lines_num += 1

                else:
                    data_type = "elf"

                # If the VASP calculation was spin polarized there will be another
                # set of data. We check this is true and then read in the data if
                # it exists
                skip_rows_diff = skip_rows_total + line_num + aug_occ_lines_num + 1
                if grid_shape_str in content[skip_rows_diff - 1]:
                    # there is a second set of data we need to load
                    for line in content[skip_rows_diff : skip_rows_diff + line_num]:
                        diff_str += f"{line}"
            except:
                # We've reached the end of the file without finding an
                # augmentation occupancies or additional set of data. We either
                # have a non-spin polarized ELF file or the output
                # BvAt_.dat file from the Henkelman Bader code.
                data_type = "unknown"

        # Create the structure object
        structure = Structure.from_str(structure_str, fmt="poscar")

        # Load in the data for the first set of data (spin up in ELFCAR, total
        # charge density in CHGCAR). Ravel moves the data from the original
        # grid in the file to a long list. Reshape moves it into a 3D numpy array
        # with the appropriate shape.
        total = (
            np.fromstring(
                total_str,
                sep=" ",
            )
            .ravel()
            .reshape(grid_shape, order="F")
        )

        # Do the same for the second set of data (spin down in ELFCAR, magnetic
        # moment in CHGCAR)
        if len(diff_str) > 0:
            diff = (
                np.fromstring(
                    diff_str,
                    sep=" ",
                )
                .ravel()
                .reshape(grid_shape, order="F")
            )
        else:
            diff = None

        return Grid(total, diff, structure, data_type)

    def write_file(self, file_name: str | Path, vasp4_compatible=False):
        """
        Writes a vasp compatible file. This is a reimplimentation of pymatgen's
        [version](https://github.com/materialsproject/pymatgen/blob/v2023.10.4/pymatgen/io/vasp/outputs.py) without the required overhead.
        Args:
            file_name (str | Path): Path to file
            vasp4_compatible: True if the format is vasp4 compatible.
        """
        logging.info(f"Writing to {file_name}")

        def _print_fortran_float(flt):
            """Fortran codes print floats with a leading zero in scientific
            notation. When writing CHGCAR files, we adopt this convention
            to ensure written CHGCAR files are byte-to-byte identical to
            their input files as far as possible.

            Args:
                flt (float): Float to print.

            Returns:
                str: String representation of float in Fortran format.
            """
            s = f"{flt:.10E}"
            if flt >= 0:
                return f"0.{s[0]}{s[2:12]}E{int(s[13:]) + 1:+03}"
            return f"-.{s[1]}{s[3:13]}E{int(s[14:]) + 1:+03}"

        with open(file_name, "wt") as file:
            poscar = Poscar(self.structure)

            # use original name if it's been set (e.g. from Chgcar)
            comment = getattr(self, "name", poscar.comment)

            lines = comment + "\n"
            lines += "   1.00000000000000\n"
            for vec in self.structure.lattice.matrix:
                lines += f" {vec[0]:12.6f}{vec[1]:12.6f}{vec[2]:12.6f}\n"
            if not vasp4_compatible:
                lines += "".join(f"{s:5}" for s in poscar.site_symbols) + "\n"
            lines += "".join(f"{x:6}" for x in poscar.natoms) + "\n"
            lines += "Direct\n"
            for site in self.structure:
                dim, b, c = site.frac_coords
                lines += f"{dim:10.6f}{b:10.6f}{c:10.6f}\n"
            lines += " \n"
            file.write(lines)
            dim = self.grid_shape

            def write_spin(data):
                lines = []
                count = 0
                file.write(f"   {dim[0]}   {dim[1]}   {dim[2]}\n")
                for k, j, i in itertools.product(
                    list(range(dim[2])), list(range(dim[1])), list(range(dim[0]))
                ):
                    lines.append(_print_fortran_float(data[i, j, k]))
                    count += 1
                    if count % 5 == 0:
                        file.write(" " + "".join(lines) + "\n")
                        lines = []
                    else:
                        lines.append(" ")
                if count % 5 != 0:
                    file.write(" " + "".join(lines) + " \n")

            write_spin(self.total)
            if self.diff is not None:
                write_spin(self.diff)

    def regrid(
        self,
        desired_resolution: int = 130000,
        new_grid_shape: np.array = None,
    ):
        """
        Resizes the grid using Fourier interplation as implemented by
        [PyRho](https://materialsproject.github.io/pyrho/)

        Args:
            desired_resolution (int): The desired resolution in voxels/A^3.
            new_grid_shape (ArrayLike): The new array shape. Takes precedence
                over desired_resolution.

        Returns:
            Changes the grid data in place.
        """
        # Get data
        total = self.total
        diff = self.diff

        # Get the lattice unit vectors as a 3x3 array
        lattice_array = self.structure.lattice.matrix

        # get the original grid size and lattice volume.
        grid_shape = self.grid_shape
        volume = self.structure.volume

        if new_grid_shape is None:
            # calculate how much the number of voxels along each unit cell must be
            # multiplied to reach the desired resolution.
            scale_factor = ((desired_resolution * volume) / grid_shape.prod()) ** (
                1 / 3
            )

            # calculate the new grid shape. round up to the nearest integer for each
            # side
            new_grid_shape = np.around(grid_shape * scale_factor).astype(np.int32)

        # create a pyrho pgrid instance.
        total_pgrid = PGrid(total, lattice_array)

        # regrid to the desired voxel resolution and get the data back out.
        new_total_pgrid = total_pgrid.get_transformed(
            [[1, 0, 0], [0, 1, 0], [0, 0, 1]], new_grid_shape
        )
        # new_data = pgrid.lossy_smooth_compression(new_grid_shape)
        new_total_data = new_total_pgrid.grid_data
        # update class total instance
        self.total = new_total_data

        # repeat for diff if exists
        if diff is not None:
            diff_pgrid = PGrid(diff, lattice_array)
            new_diff_pgrid = diff_pgrid.get_transformed(
                [[1, 0, 0], [0, 1, 0], [0, 0, 1]], new_grid_shape
            )
            new_diff_data = new_diff_pgrid.grid_data
            self.diff = new_diff_data

    ###########################################################################
    # The following is a series of methods that are useful for converting between
    # voxel coordinates, fractional coordinates, and cartesian coordinates.
    ###########################################################################
    def get_voxel_coords_from_index(self, site):
        """
        Takes in a site index and returns the equivalent voxel grid index.

        Args:
            site (int): the index of the site to find the grid index for

        Returns:
            A voxel grid index as an array.

        """

        voxel_coords = [
            a * b + 1 for a, b in zip(self.grid_shape, self.frac_coords[site])
        ]
        # voxel positions go from 1 to (grid_size + 0.9999)
        return np.array(voxel_coords)

    def get_voxel_coords_from_neigh_CrystalNN(self, neigh):
        """
        Gets the voxel grid index from a neighbor atom object from CrystalNN or
        VoronoiNN

        Args:
            neigh (Neigh): a neighbor type object from pymatgen

        Returns:
            A voxel grid index as an array.
        """
        grid_size = self.grid_shape
        frac = neigh["site"].frac_coords
        voxel_coords = [a * b + 1 for a, b in zip(grid_size, frac)]
        # voxel positions go from 1 to (grid_size + 0.9999)
        return np.array(voxel_coords)

    def get_voxel_coords_from_neigh(self, neigh):
        """
        Gets the voxel grid index from a neighbor atom object from the pymatgen
        structure.get_neighbors class.

        Args:
            neigh (dict): a neighbor dictionary from pymatgens structure.get_neighbors
                method.

        Returns:
            A voxel grid index as an array.
        """
        grid_size = self.grid_shape
        frac_coords = neigh.frac_coords
        voxel_coords = [a * b + 1 for a, b in zip(grid_size, frac_coords)]
        # voxel positions go from 1 to (grid_size + 0.9999)
        return np.array(voxel_coords)

    def get_voxel_coords_from_frac(self, frac_coords: ArrayLike | list):
        """
        Takes in a fractional coordinate and returns the cartesian coordinate.

        Args:
            frac_coords (ArrayLike): The fractional position to convert to cartesian coords.

        Returns:
            A voxel grid index as an array.

        """
        grid_size = self.grid_shape
        voxel_coords = [a * b + 1 for a, b in zip(grid_size, frac_coords)]
        # voxel positions go from 1 to (grid_size + 0.9999)
        return np.array(voxel_coords)

    def get_frac_coords_from_vox(self, voxel_coordsition: ArrayLike | list):
        """
        Takes in a voxel grid index and returns the fractional
        coordinates.

        Args:
            voxel_coordsition (ArrayLike): A voxel grid index

        Returns:
            A fractional coordinate as an array
        """
        size = self.grid_shape
        fa, fb, fc = [((a - 1) / b) for a, b in zip(voxel_coordsition, size)]
        frac_coords = np.array([fa, fb, fc])
        return frac_coords

    def get_cart_coords_from_frac(self, frac_coords: ArrayLike | list):
        """
        Takes in fractional coordinates and returns cartesian coordinates

        Args:
            frac_coords (ArrayLike): The fractional position to convert to cartesian coords.

        Returns:
            Cartesian coordinates as an array
        """
        fa, fb, fc = frac_coords[0], frac_coords[1], frac_coords[2]
        a, b, c = self.a, self.b, self.c
        x = fa * a[0] + fb * b[0] + fc * c[0]
        y = fa * a[1] + fb * b[1] + fc * c[1]
        z = fa * a[2] + fb * b[2] + fc * c[2]
        cart_coords = np.array([x, y, z])
        return cart_coords

    def get_frac_coords_from_cart(self, cart_coords: ArrayLike | list):
        """
        Takes in a cartesian coordinate and returns the fractional coordinates.

        Args:
            cart_coords (ArrayLike): A cartesian coordinate.

        Returns:
            fractional coordinates as an Array
        """
        lattice_matrix = self.structure.lattice.matrix
        inverse_matrix = np.linalg.inv(lattice_matrix)
        frac_coords = np.dot(cart_coords, inverse_matrix)

        return frac_coords

    def get_cart_coords_from_vox(self, voxel_coord: ArrayLike | list):
        """
        Takes in a voxel grid index and returns the cartesian coordinates.

        Args:
            voxel_coords (ArrayLike): A voxel grid index

        Returns:
            Cartesian coordinates as an array
        """
        frac_coords = self.get_frac_coords_from_vox(voxel_coord)
        cart_coords = self.get_cart_coords_from_frac(frac_coords)
        return cart_coords

    def get_voxel_coords_from_cart(self, cart_coords: ArrayLike | list):
        """
        Takes in a cartesian coordinate and returns the voxel coordinates.

        Args:
            cart_coords (ArrayLike): A cartesian coordinate.

        Returns:
            Voxel coordinates as an Array
        """
        frac_coords = self.get_frac_coords_from_cart(cart_coords)
        voxel_coords = self.get_voxel_coords_from_frac(frac_coords)
        return voxel_coords

    def _plot_points(self, points, ax, fig, color, size: int = 20):
        """
        Plots points of form [x,y,z] using matplotlib

        Args:
            points (list): A list of points to plot
            fig: A matplotlib.pyplot.figure() instance
            ax: A matplotlib Subplot instance
            color (str): The color to plot the points
            size (int): The pt size to plot
        """
        x = []
        y = []
        z = []
        for point in points:
            x.append(point[0])
            y.append(point[1])
            z.append(point[2])
        ax.scatter(x, y, z, c=color, s=size)

    def _plot_unit_cell(self, ax, fig):
        """
        Plots the unit cell of a structure using matplotlib

        Args:
            fig: A matplotlib.pyplot.figure() instance
            ax: A matplotlib Subplot instance
        """
        if ax is None or fig is None:
            fig = plt.figure()
            ax = fig.add_subplot(projection="3d")

        # Get the points at the lattice vertices to plot and plot them
        a = self.a
        b = self.b
        c = self.c
        points = [np.array([0, 0, 0]), a, b, c, a + b, a + c, b + c, a + b + c]
        self._plot_points(points, ax, fig, "purple")

        # get the structure to plot.
        structure = self.structure
        species = structure.symbol_set

        # get a color map to distinguish between sites
        color_map = matplotlib.colormaps.get_cmap("tab10")
        # Go through each atom type and plot all instances with the same color
        for i, specie in enumerate(species):
            color = color_map(i)
            sites_indices = structure.indices_from_symbol(specie)
            for site in sites_indices:
                coords = structure[site].coords
                self._plot_points([coords], ax, fig, color, size=40)



# -*- coding: utf-8 -*-

from simmate.apps.warren_lab.workflows.relaxation.scan import (
    Relaxation__Vasp__WarrenLabScan,
)
from simmate.apps.warren_lab.workflows.static_energy.hse import HSE_STATIC_SETTINGS


class StaticEnergy__Vasp__WarrenLabScan(Relaxation__Vasp__WarrenLabScan):
    """
    Performs a static energy calculation based on the settings for Warren Lab
    SCAN functional relaxation.
    """

    _incar_updates = HSE_STATIC_SETTINGS



# -*- coding: utf-8 -*-

import csv
import itertools
import logging
import math
import warnings
from functools import cached_property
from pathlib import Path

import dask.dataframe
import numpy as np
import pandas as pd
import psutil
from dask.distributed import Client, LocalCluster
from pymatgen.analysis.dimensionality import get_dimensionality_larsen
from pymatgen.analysis.graphs import StructureGraph
from pymatgen.analysis.local_env import CrystalNN
from pymatgen.io.vasp import Potcar

from simmate.apps.badelf.core.electride_finder import ElectrideFinder
from simmate.apps.badelf.core.grid import Grid
from simmate.apps.badelf.core.partitioning import PartitioningToolkit
from simmate.apps.badelf.core.voxel_assignment import VoxelAssignmentToolkit
from simmate.apps.bader.outputs import ACF
from simmate.workflows.utilities import get_workflow

# BUG: we shouldn't fully turning off warnings. This should be used within a context.
warnings.filterwarnings("ignore")


class BadElfToolkit:
    """
    A set of tools for performing BadELF, VoronELF, or Zero-Flux analysis on
    outputs from a VASP calculation.

    Args:
        partitioning_grid (Grid):
            A badelf app Grid like object used for partitioning the unit cell
            volume. Usually contains ELF.
        charge_grid (Grid):
            A badelf app Grid like object used for summing charge. Usually
            contains charge density.
        directory (Path):
            The Path to perform the analysis in.
        cores (int):
            The number of cores (NOT threads) to use for voxel assignment.
            Defaults to 0.9*the total number of cores available.
        algorithm (str):
            The algorithm to use for partitioning. Options are "badelf", "voronelf",
            or "zero-flux".
        find_electrides (bool):
            Whether or not to search for electride sites. Usually set to true.
    """

    check_for_covalency = True
    electride_finder_cutoff = 0.5
    electride_connection_cutoff = 0

    def __init__(
        self,
        partitioning_grid: Grid,
        charge_grid: Grid,
        directory: Path,
        cores: int = None,
        algorithm: str = "badelf",
        find_electrides: bool = True,
    ):
        if partitioning_grid.structure != charge_grid.structure:
            raise ValueError("Grid structures must be the same.")
        if cores is None:
            self.cores = math.floor(len(psutil.Process().cpu_affinity()) * 0.9 / 2)
        else:
            self.cores = cores

        # Check if POTCAR exists in path. If not, throw warning
        if not (directory / "POTCAR").exists():
            raise Exception(
                """
                No POTCAR file found in the requested directory.
                """
            )
        if algorithm not in ["badelf", "voronelf", "zero-flux"]:
            raise ValueError(
                """The algorithm setting you chose does not exist. Please select
                  either 'badelf', 'voronelf', or 'zero-flux'.
                  """
            )

        self.partitioning_grid = partitioning_grid
        self.charge_grid = charge_grid
        self.directory = directory
        self.algorithm = algorithm
        self.find_electrides = find_electrides

        self._voxel_errors = None  # Stores voxel coordinates that had errors when
        # Being assigned. Only filled after badelf/vorelf have run

    @cached_property
    def structure(self):
        structure = self.partitioning_grid.structure.copy()
        structure.remove_species(["He"])
        return structure

    @cached_property
    def electride_structure(self):
        """
        Searches the partitioning grid for potential electride sites and returns
        a structure with the found sites.

        Returns:
            A Structure object with electride sites as "He" atoms.
        """

        if self.find_electrides:
            electride_structure = ElectrideFinder(
                self.partitioning_grid,
                self.directory,
            ).get_electride_structure(
                electride_finder_cutoff=self.electride_finder_cutoff
            )
        else:
            electride_structure = self.structure

        return electride_structure

    @cached_property
    def electride_indices(self):
        """
        The indices of the structure that are electride sites.
        """
        return self.electride_structure.indices_from_symbol("He")

    @cached_property
    def coord_envs(self):
        """
        The coordination environment around each electride.
        """
        return self._get_coord_envs()

    def _get_coord_envs(self):
        """
        Gets the coordination environment for electrides in the system.

        Returns:
            A list of coordination environments in order of the electrides in
            the system.
        """
        # create a CrystalNN loop
        cnn = CrystalNN()
        coord_envs = []
        # For each electride in the structure, we make a new temporary structure
        # without electride sites. We then add one electride site at a time and
        # check its environment.
        for i, site in enumerate(self.electride_structure):
            coord_envs.append(cnn.get_cn(structure=self.electride_structure, n=i))
        return coord_envs

    @cached_property
    def partitioning(self):
        """
        The partitioning planes for the structure as a dictionary of dataframes.
        None if the zero-flux method is selected
        """
        return self._get_partitioning()

    def _get_partitioning(self):
        """
        Gets the partitioning used in the badelf and voronelf algorithms.

        Returns:
            Dictionary relating sites to the planes surrounding them. None
            if the zero-flux algorithm is selected.
        """
        # Get the partitioning grid
        partitioning_grid = self.partitioning_grid.copy()
        # partitioning_grid.regrid()
        # If the algorithm is badelf, we don't want to partition with the structure
        # containing electrides. We remove any electrides in case the provided
        # structure already had them.
        if self.algorithm == "badelf":
            # remove electrides from grid structure and get
            partitioning_grid.structure.remove_species("He")
            partitioning = PartitioningToolkit(partitioning_grid).get_partitioning(
                check_for_covalency=self.check_for_covalency
            )
            return partitioning
        elif self.algorithm == "voronelf":
            # Use the structure with electrides as the partitioning structure.
            # This will not be anything different from the base structure if there
            # are no electride sites.
            partitioning_grid.structure = self.electride_structure.copy()
            partitioning = PartitioningToolkit(partitioning_grid).get_partitioning(
                check_for_covalency=self.check_for_covalency
            )
            return partitioning
        elif self.algorithm == "zero-flux":
            print(
                """
                There is no partitioning property for the zero-flux algorithm as
                the partitioning is handled by the [Henkelman Bader code](https://theory.cm.utexas.edu/henkelman/code/bader/)
                """
            )
            return None

    @cached_property
    def voxel_assignments(self):
        """
        A dataframe with each voxel coordinate, charge, and dictionary of sites
        that the voxel is assigned to.
        """
        return self._get_voxel_assignments()

    def _get_voxel_assignments(self):
        """
        Gets a dataframe of voxel assignments. The dataframe has columns
        [x, y, z, charge, sites]
        """
        logging.info("Beginning voxel assignment (this can take a while)")
        algorithm = self.algorithm
        if algorithm == "zero-flux":
            print(
                """
                There is no voxel assignment for the zero-flux algorithm as
                the assignment is handled by the [Henkelman Bader code](https://theory.cm.utexas.edu/henkelman/code/bader/)
                """
            )
            return None

        # Get the objects that we'll need to assign voxels.
        elif algorithm in ["badelf", "voronelf"]:
            a, b, c = self.charge_grid.grid_shape
            charge_data = self.charge_grid.total

            # Create lists that contain the coordinates of each voxel and their charges
            voxel_coords = [
                idx for idx in itertools.product(range(a), range(b), range(c))
            ]
            voxel_charges = [
                float(charge_data[idx[0], idx[1], idx[2]]) for idx in voxel_coords
            ]

            # Create a dataframe that has each coordinate index and the charge as columns
            # Later we'll remove voxels that belong to electrides from this dataframe
            voxel_assignments = pd.DataFrame(voxel_coords, columns=["x", "y", "z"]).add(
                1
            )
            # create columns for site and charge.
            voxel_assignments["charge"] = voxel_charges
            voxel_assignments["site"] = None

            # If we are running BadELF we need to get the voxels assigned to electrides
            # by the henkelman bader algorithm
            if algorithm == "badelf":
                voxel_assignments = self._get_zero_flux_electride_assignment(
                    voxel_assignments
                )

            voxel_assignments = self._get_plane_seperated_voxels_assignment(
                voxel_assignments
            )

            # get voxel assignment errors
            vert_multi_site_same_trans = []
            vert_multi_site = []
            multi_site_no_plane = []
            error_indices = []
            for row in voxel_assignments.iterrows():
                # Check each voxels site assignment (stored in row[1][4]). If
                # -1, the vertices of the voxel were found to potentially belong
                # to multiple sites even at the same transformation. If -2, they
                # were found to potentially belong to multiple sites at different
                # transformations. If -3, the vertices were found to have different
                # sites, but no plane was found to intersect the voxel.

                if row[1][4] is not None:
                    for site in row[1][4].keys():
                        if site == -1:
                            vert_multi_site_same_trans.append(
                                [row[1][0], row[1][1], row[1][2]]
                            )
                            error_indices.append(row[0])
                        elif site == -2:
                            vert_multi_site.append([row[1][0], row[1][1], row[1][2]])
                            error_indices.append(row[0])
                        elif site == -3:
                            multi_site_no_plane.append(
                                [row[1][0], row[1][1], row[1][2]]
                            )
                            error_indices.append(row[0])
            self._voxel_errors = {
                "vert_multi_site_same_trans": vert_multi_site_same_trans,
                "vert_multi_site": vert_multi_site,
                "multi_site_no_plane": multi_site_no_plane,
            }
            self._write_voxel_errors()

            # Remove the site errors
            voxel_assignments["site"] = (
                np.where(  # assigns values based on boolean value. returns array
                    voxel_assignments.index.isin(
                        error_indices
                    ),  # Is True where df index is in indices list
                    None,  # The value to be assigned
                    voxel_assignments["site"],
                )
            )  # Value to assign where condition is False

            voxel_assignments = self._get_multi_plane_voxel_assignment(
                voxel_assignments
            )
        logging.info("Finished voxel assignment")
        return voxel_assignments

    def write_electride_structure_files(
        self,
        charge_file: str = "CHGCAR_w_empty_atoms",
        partitioning_file: str = "ELFCAR_w_empty_atoms",
    ):
        """
        Writes copies of the charge file and partitioning file (usually CHGCAR
        and ELFCAR) with electride sites. This is most frequently used for
        eventually running the Henkelman Bader code to get electride charges.
        """
        # Write CHGCAR and ELFCAR files with the empty structure that was found
        electride_charge_grid = self.charge_grid.copy()
        electride_charge_grid.structure = self.electride_structure
        electride_charge_grid.write_file(charge_file)
        electride_elf_grid = self.partitioning_grid.copy()
        # check that elf grid is same size as charge grid and if not, regrid
        if electride_charge_grid.voxel_num != electride_elf_grid.voxel_num:
            electride_elf_grid.regrid(new_grid_shape=electride_charge_grid.grid_shape)
        electride_elf_grid.structure = self.electride_structure
        electride_elf_grid.write_file(partitioning_file)

    def _get_zero_flux_electride_assignment(self, voxel_assignments: pd.DataFrame):
        """
        Gets the electride site assignments for voxels belonging to electride
        sites based on the Henkelman groups algorithm.

        Args:
            voxel_assignments (DataFrame):
                The dataframe to add the voxel assignments to.

        Returns:
            updated voxel_assignments dataframe
        """

        directory = self.directory
        charge_file = directory / "CHGCAR_w_empty_atoms"
        partitioning_file = directory / "ELFCAR_w_empty_atoms"
        if not (directory / "CHGCAR_w_empty_atoms").exists():
            self.write_electride_structure_files(charge_file, partitioning_file)
        # Run the henkelman code to print out electride files
        badelf_workflow = get_workflow("population-analysis.bader.bader-dev")
        badelf_workflow.run(
            directory=directory,
            charge_file="CHGCAR_w_empty_atoms",
            partitioning_file="ELFCAR_w_empty_atoms",
            atoms_to_print=self.electride_indices,
        )

        for electride in self.electride_indices:
            # Create a dictionary with all sites to store the value in. We do
            # this to have the same format for all sites including those that
            # are shared by more than one atom
            site_vol_frac = {}
            for i, site in enumerate(self.electride_structure):
                site_vol_frac[i] = float(0)
            site_vol_frac[electride] = float(1)
            # Pull in electride charge density from bader output file (BvAt####.dat format)
            self._fix_BvAt(f"BvAt{str(electride+1).zfill(4)}.dat")
            electride_charge = Grid.from_file(
                directory / f"BvAt{str(electride+1).zfill(4)}.dat"
            ).total
            electride_indices = []
            # For each voxel, check if the electride charge density file has any
            # charge. If it does add its index to the electride_indices list
            for index, row in voxel_assignments.iterrows():
                charge_density = float(
                    electride_charge[row["x"] - 1, row["y"] - 1, row["z"] - 1]
                )
                if charge_density != 0:
                    # get indices of electride sites
                    electride_indices.append(index)
                    # results_charge[electride] += charge_density
            # Update all of the electride indices with the site dictionary
            voxel_assignments["site"] = (
                np.where(  # assigns values based on boolean value. returns array
                    voxel_assignments.index.isin(
                        electride_indices
                    ),  # Is True where df index is in indices list
                    site_vol_frac,  # The value to be assigned
                    voxel_assignments["site"],
                )
            )  # Value to assign where condition is False

        return voxel_assignments

    def _get_plane_seperated_voxels_assignment(
        self,
        voxel_assignments: pd.DataFrame,
    ):
        """
        Gets the electride site assignments for voxels based on partitioning.

        Args:
            voxel_assignments (DataFrame):
                The dataframe to add the voxel assignments to.

        Returns:
            updated voxel_assignments dataframe
        """
        voxel_assignment_tools = VoxelAssignmentToolkit(
            charge_grid=self.charge_grid,
            partitioning_grid=self.partitioning_grid,
            partitioning=self.partitioning,
            electride_structure=self.electride_structure,
            algorithm=self.algorithm,
        )
        # partitioning = self.partitioning
        # permutations = self.charge_grid.permutations
        cores = self.cores
        # get the amount of memory available
        memory_gb = psutil.virtual_memory()[1] / (1e9)
        # Each worker needs at least 2GB of memory. We select either the number
        # of workers that could have at least this much memory or the number
        # of cores/threads, whichever is smaller
        nworkers = min(math.floor(memory_gb / 2), cores)

        # We need the total number of voxels so that we can make sure our partitions
        # aren't too large. I did some light testing and found that 128000 was a
        # good limit for the number of voxels that could be handled by a worker
        # with 2GB of memory. Much larger and it could start having issues.
        total_voxels = len(voxel_assignments)
        if total_voxels <= 128000 * nworkers:
            npartitions = nworkers
        elif total_voxels > 128000 * nworkers:
            npartitions = math.ceil(total_voxels / 128000)

        with (
            LocalCluster(
                n_workers=nworkers,
                threads_per_worker=2,
                memory_limit="auto",
                processes=True,
            ) as cluster,
            Client(cluster) as client,
        ):
            # put list of indices in dask dataframe. Partition with the same
            # number of partitions as workers
            ddf = dask.dataframe.from_pandas(
                voxel_assignments,
                npartitions=npartitions,
            )

            # site search for all voxel positions.
            ddf["site"] = ddf.map_partitions(
                voxel_assignment_tools._get_voxels_site_dask,
            )
            # run site search and save as pandas dataframe
            pdf = ddf.compute()

        return pdf

    def _get_multi_plane_voxel_assignment(
        self,
        voxel_assignments: pd.DataFrame,
    ):
        voxel_assignment_tools = VoxelAssignmentToolkit(
            charge_grid=self.charge_grid,
            partitioning_grid=self.partitioning_grid,
            partitioning=self.partitioning,
            electride_structure=self.electride_structure,
            algorithm=self.algorithm,
        )
        voxel_assignments["site"] = voxel_assignments.apply(
            lambda x: voxel_assignment_tools.get_voxels_multi_plane(
                point_voxel_coords=[x["x"], x["y"], x["z"]],
                site_vol_frac=x["site"],
                voxel_assignments=voxel_assignments,
            ),
            axis=1,
        )
        return voxel_assignments

    def _write_voxel_errors(self):
        """
        Writes any voxel errors that were found to a csv file.
        """

        dataframe = pd.DataFrame.from_dict(
            dict([(key, pd.Series(value)) for key, value in self._voxel_errors.items()])
        )
        dataframe.to_csv(self.directory / "same_site_voxels.csv")

    @cached_property
    def voxel_assignments_array(self):
        """
        An array that relates each voxel in the charge_grid to a site.
        """
        return self._get_voxel_assignments_array()

    def _get_voxel_assignments_array(self):
        voxel_assignments = self.voxel_assignments

        # Make blank array
        voxel_array = np.zeros_like(self.charge_grid.total)
        for row in voxel_assignments.iterrows():
            sites = row[1][4]
            max_site_frac = max(sites.values())
            max_site = list(sites.keys())[list(sites.values()).index(max_site_frac)]
            voxel_array[row[1][0] - 1, row[1][1] - 1, row[1][2] - 1] = max_site
        return voxel_array

    def get_electride_dimensionality(self, electride_connection_cutoff: float = 0):
        electride_indices = self.electride_indices
        # If we have no electrides theres no reason to continue so we stop here
        if len(electride_indices) == 0:
            return None

        if self.algorithm == "zero-flux":
            # !!! read in electride only ELFCAR. Regrid to charge_grid size
            # Get the necessary CHGCAR and ELFCAR files. Then run the
            # Henkelman Bader code, printing the resulting electride voxels in
            # one file for topolgoy analysis.
            directory = self.directory
            partitioning_file = directory / "ELFCAR_w_empty_atoms"
            if not partitioning_file.exists():
                self.write_electride_structure_files(
                    directory / "CHGCAR_w_empty_atoms", partitioning_file
                )
            badelf_workflow = get_workflow("population-analysis.bader.bader-dev")
            badelf_workflow.run(
                directory=directory,
                charge_file="ELFCAR_w_empty_atoms",
                partitioning_file="ELFCAR_w_empty_atoms",
                species_to_print="He",
                structure=self.electride_structure,
            )
            self._fix_BvAt("BvAt_summed.dat")
            elf_grid = Grid.from_file(directory / "BvAt_summed.dat")
            elf_grid.regrid(desired_resolution=self.charge_grid.voxel_resolution)
            pass
        elif self.algorithm in ["badelf", "voronelf"]:
            # read in ELF data and regrid so that it is the same size as the
            # charge grid
            elf_grid = self.partitioning_grid.copy()
            elf_grid.regrid(desired_resolution=self.charge_grid.voxel_resolution)
            voxel_assignment_array = self.voxel_assignments_array
            # Get array where values are ELF values when voxels belong to electrides
            # and are 0 otherwise
            elf_array = np.where(
                np.isin(voxel_assignment_array, electride_indices), elf_grid.total, 0
            )
            elf_grid.total = elf_array

        # read in structure and remove all atoms except dummy electride sites
        electride_structure = self.electride_structure.copy()

        electride_structure.remove_species(self.structure.symbol_set)

        elf_grid.structure = electride_structure

        partitioning_tools = PartitioningToolkit(elf_grid)

        # get the 50 nearest electride neighbors. We only do this because we need to make
        # sure that electride sites that are very far away are thoroughly checked
        nearest_neighbors = partitioning_tools.get_set_number_of_neighbors(50)

        # create an empty StructureGraph object. This maps connections between different
        # atoms, including those across unit cell boundaries. We will fill this out using
        # the list of neighbors we just defined.
        graph = StructureGraph.with_empty_graph(electride_structure)
        # partitioning_lines = []
        # iterate over each unique electride site.
        for index, neighbors in enumerate(nearest_neighbors):
            # get the voxel coords for the electride site
            site_pos = elf_grid.get_voxel_coords_from_index(index)

            # loop over the neighboring electride sites
            for neighbor in neighbors:
                # We will have many excess electride sites that are in unit cells that
                # don't border the one we're looking at. I'm not certain, but I don't
                # think those are necessary. We cut them out here to save time from the
                # get_partitioning_line_from_voxels function.

                # Assume this neighbor is not more than one unit cell away
                more_than_1_unit_cell_away = False
                for integer in neighbor.image:
                    # Check if an integer other than -1, 1, or 0 is present. This indicates
                    # we've been transformed more than one unit cell away.
                    if integer not in [-1, 0, 1]:
                        more_than_1_unit_cell_away = True
                        break
                # Check if we've moved more than one unit cell away. If we have, skip
                # to the next neighbor
                if more_than_1_unit_cell_away:
                    continue

                # get the voxel coord for the connected electride site and get the ELF
                # line between the site and this neighbor
                neigh_pos = elf_grid.get_voxel_coords_from_neigh(neighbor)
                pos, values = partitioning_tools.get_partitioning_line_from_voxels(
                    site_pos, neigh_pos
                )
                # partitioning_lines.append([pos,values])
                # If a 0 is not found in the elf line these sites are connected and we
                # want to add an edge to our graph.
                #!!! create cutoff for how much space is allowed between sites
                # for when vorelf cuts it off by a voxel or so?
                if all(value > electride_connection_cutoff for value in values):
                    graph.add_edge(
                        from_index=index,  # The site index of the electride site of interest
                        from_jimage=(
                            0,
                            0,
                            0,
                        ),  # The image the electride site is in. Always (0,0,0)
                        to_index=neighbor.index,  # The neighboring electrides site index
                        to_jimage=neighbor.image,  # The image that the neighbor is in.
                        weight=None,  # The relative weight of the neighbor. We ignore this.
                        edge_properties=None,
                        warn_duplicates=False,  # Duplicates are fine for us.
                    )

        # Get the dimensionality from our StructureGraph. If more than one group of electrides
        # is found, it will default to the highest dimensionality.
        return get_dimensionality_larsen(graph)

    def _fix_BvAt(self, file_name):
        electride_structure = self.electride_structure
        symbols = electride_structure.types_of_species
        new_symbol_line = ""
        for symbol in symbols:
            new_symbol_line += f"{symbol.name}   "
        directory = self.directory
        with open(directory / f"{file_name}", "r") as file:
            content = file.readlines()
        symbol_line = content[5]
        if not all(symbol.name in symbol_line for symbol in symbols):
            content[5] = new_symbol_line + "\n"
            with open(directory / f"{file_name}", "w") as file:
                file.writelines(content)

    @cached_property
    def results(self):
        """
        A summary of the results from a BadELF run.
        """
        return self._get_results()

    def _get_results(self):
        algorithm = self.algorithm
        directory = self.directory
        electride_num = len(self.electride_indices)
        electride_structure = self.electride_structure
        electride_indices = self.electride_indices
        a, b, c = self.charge_grid.grid_shape
        elements = []
        for site in electride_structure:
            if site.species_string == "He":
                elements.append("e")
            else:
                elements.append(site.species_string)

        if algorithm == "zero-flux":
            # Get the necessary CHGCAR and ELFCAR files. Then run the
            # Henkelman Bader code, printing the resulting electride voxels in
            # one file for topolgoy analysis.
            charge_file = directory / "CHGCAR_w_empty_atoms"
            partitioning_file = directory / "ELFCAR_w_empty_atoms"
            self.write_electride_structure_files(charge_file, partitioning_file)
            badelf_workflow = get_workflow("population-analysis.bader.bader-dev")
            badelf_workflow.run(
                directory=directory,
                charge_file="CHGCAR_w_empty_atoms",
                partitioning_file="ELFCAR_w_empty_atoms",
            )
            # get the desired data that will be saved to the dataframe
            #!!! I should rework the ACF.dat reader now that I have better
            # tools than pymatgen's reader
            logging.info("Calculating additional useful information")
            results_dataframe, extra_data = ACF(directory)
            results = {
                "oxidation_states": list(results_dataframe.oxidation_state.values),
                "charges": list(results_dataframe.charge.values),
                "min_dists": list(results_dataframe.min_dist.values),
                "atomic_volumes": list(results_dataframe.atomic_vol.values),
                **extra_data,
            }

        elif algorithm in ["badelf", "voronelf"]:
            # get the voxel assignments as a dataframe.
            voxel_assignments = self.voxel_assignments
            voxel_volume = self.charge_grid.voxel_volume
            logging.info("Calculating additional useful information")
            # Create dictionaries to store the important results
            min_dists = {}
            charges = {}
            atomic_volumes = {}
            for site in range(len(electride_structure)):
                charges[site] = 0
                atomic_volumes[site] = 0
            # Get the minimum distances from each atom the the partitioning
            # surface. If the algorithm is "badelf" we need to acquire the
            # radii for the electrides
            if algorithm == "badelf":
                results_dataframe, extra_data = ACF(directory)
                electride_min_dists = results_dataframe.min_dist
            for site in range(len(electride_structure)):
                # fill min_dist dictionary using the smallest partitioning radius
                if site in electride_indices and algorithm == "badelf":
                    # Get dist from henkelman algorithm results
                    min_dists[site] = electride_min_dists[site]
                else:
                    # Get dists from partitioning
                    radii = []
                    for neighbor_index, row in self.partitioning[site].iterrows():
                        radii.append(row["radius"])
                    min_radii = min(radii)
                    min_dists[site] = min_radii

            # Get the charge and atomic volume of each site.
            for row in voxel_assignments.iterrows():
                for site in row[1][4]:
                    charges[site] += row[1][4][site] * row[1][3]
                    atomic_volumes[site] += row[1][4][site] * voxel_volume
            # Convert charges from VASP standard
            for site, charge in charges.items():
                charges[site] = charge / (a * b * c)

            # Get the number of electrons assigned by badelf.
            nelectrons = round(sum(charges.values()), 6)

            # Get
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                potcars = Potcar.from_file(directory / "POTCAR")
            nelectron_data = {}
            # the result is a list because there can be multiple element potcars
            # in the file (e.g. for NaCl, POTCAR = POTCAR_Na + POTCAR_Cl)
            for potcar in potcars:
                nelectron_data.update({potcar.element: potcar.nelectrons})

            # create lists to store the element list, oxidation states, charges,
            # minimum distances, and atomic volumes
            oxi_state_data = []
            charges_list = []
            min_dists_list = []
            atomic_volumes_list = []
            # iterate over the charge results and add the results to each list
            for site_index, site_charge in charges.items():
                # get structure site
                site = electride_structure[site_index]
                # get element name
                element_str = site.specie.name
                # Change electride dummy atom name to e
                if element_str == "He":
                    element_str = "e"
                # calculate oxidation state and add it to the oxidation state list
                if element_str == "e":
                    oxi_state = -site_charge
                else:
                    oxi_state = nelectron_data[element_str] - site_charge
                oxi_state_data.append(oxi_state)
                # add the corresponding charge, distance, and atomic volume to the
                # respective lits
                charges_list.append(site_charge)
                min_dists_list.append(min_dists[site_index])
                atomic_volumes_list.append(atomic_volumes[site_index])

            # Calculate the "vacuum charge" or the charge not associated with any atom.
            # Idealy this should be 0.
            total_electrons = 0
            for site in self.structure:
                site_valence_electrons = nelectron_data[site.species_string]
                total_electrons += site_valence_electrons
            vacuum_charge = round((total_electrons - nelectrons), 6)

            # Calculate the "vacuum volume" or the volume not associated with any atom.
            # Idealy this should be 0
            vacuum_volume = round(
                (self.structure.volume - sum(atomic_volumes.values())), 6
            )

            # Save everything in a results dictionary
            results = {
                "oxidation_states": oxi_state_data,
                "charges": charges_list,
                "min_dists": min_dists_list,
                "atomic_volumes": atomic_volumes_list,
                "vacuum_charge": vacuum_charge,
                "vacuum_volume": vacuum_volume,
                "nelectrons": nelectrons,
            }

        # set the results that are not algorithm dependent
        results["nelectrides"] = electride_num
        results["algorithm"] = algorithm
        results["element_list"] = elements
        results["coord_envs"] = self.coord_envs
        results["electride_dim"] = self.get_electride_dimensionality(
            electride_connection_cutoff=self.electride_connection_cutoff
        )
        results["elf_connect_cutoff"] = self.electride_connection_cutoff
        # Fill out columns unrelated to badelf alg
        structure = self.structure
        results["structure"] = structure

        return results

    def write_results_csv(self):
        directory = self.directory
        results = self.results
        with open(directory / "badelf_summary.csv", "w") as csv_file:
            writer = csv.writer(csv_file)
            for key, value in results.items():
                writer.writerow([key, value])

    @classmethod
    def from_files(
        cls,
        directory: Path = Path("."),
        partitioning_file: str = "ELFCAR",
        charge_file: str = "CHGCAR",
        algorithm: str = "badelf",
        find_electrides: bool = True,
        cores: int = None,
    ):
        """
        Creates a BadElfToolkit instance from the requested partitioning file
        and charge file.

        Args:
            directory (Path):
                The path to the directory that the badelf analysis
                will be located in.
            partitioning_file (str):
                The filename of the file to use for partitioning. Must be a VASP
                CHGCAR or ELFCAR type file.
            charge_file (str):
                The filename of the file containing the charge information. Must
                be a VASP CHGCAR or ELFCAR type file.
            algorithm (str):
                The algorithm to use. Options are "badelf", "voronelf", or "zero-flux"
            find_electrides (bool):
                Whether or not to search for electrides in the system
            cores (int):
                The number of computer cores (NOT threads) that will be used to
                parallelize
        """

        partitioning_grid = Grid.from_file(directory / partitioning_file)
        charge_grid = Grid.from_file(directory / charge_file)
        return BadElfToolkit(
            partitioning_grid=partitioning_grid,
            charge_grid=charge_grid,
            directory=directory,
            algorithm=algorithm,
            find_electrides=find_electrides,
        )

    def write_species_file(self, file_type: str = "elf", species: str = "He"):
        voxel_assignment_array = self.voxel_assignments_array
        if file_type == "elf":
            grid = self.partitioning_grid.copy()
            grid.regrid(desired_resolution=self.charge_grid.voxel_resolution)
        elif file_type == "charge":
            grid = self.charge_grid.copy()
        else:
            raise ValueError(
                """
                Invalid file_type. Options are "elf" or "charge".
                """
            )
        grid.structure = self.electride_structure
        indices = self.electride_structure.indices_from_symbol(species)
        # Get array where values are ELF values when voxels belong to electrides
        # and are 0 otherwise
        array = np.where(np.isin(voxel_assignment_array, indices), grid.total, 0)
        grid.total = array
        if grid.diff is not None:
            diff_array = np.where(
                np.isin(voxel_assignment_array, indices), grid.diff, 0
            )
            grid.diff = diff_array

        if species == "He":
            species = "e"
        if file_type == "elf":
            grid.write_file(f"ELFCAR_{species}")
        elif file_type == "charge":
            grid.write_file(f"CHGCAR_{species}")

    def write_atom_file(
        self,
        atom_index: int,
        file_type: str = "elf",
    ):
        voxel_assignment_array = self.voxel_assignments_array
        if file_type == "elf":
            grid = self.partitioning_grid.copy()
            grid.regrid(desired_resolution=self.charge_grid.voxel_resolution)
        elif file_type == "charge":
            grid = self.charge_grid.copy()
        else:
            raise ValueError(
                """
                Invalid file_type. Options are "elf" or "charge".
                """
            )
        grid.structure = self.electride_structure
        # Get array where values are ELF values when voxels belong to electrides
        # and are 0 otherwise
        array = np.where(np.isin(voxel_assignment_array, atom_index), grid.total, 0)
        grid.total = array
        if grid.diff is not None:
            diff_array = np.where(
                np.isin(voxel_assignment_array, atom_index), grid.diff, 0
            )
            grid.diff = diff_array

        if file_type == "elf":
            grid.write_file(f"ELFCAR_{atom_index}")
        elif file_type == "charge":
            grid.write_file(f"CHGCAR_{atom_index}")

    def plot_partitioning(self):
        """
        Plots the partitioning surface around each atom.
        """
        partitioning = self.partitioning
        grid = self.partitioning_grid.copy()
        if self.algorithm == "badelf":
            grid.structure = self.structure
            PartitioningToolkit(grid).plot_partitioning_results(partitioning)
        elif self.algorithm == "voronelf":
            grid.structure = self.electride_structure
            PartitioningToolkit(grid).plot_partitioning_results(partitioning)
        else:
            print(
                f"""
                Plotting of zero-flux partitioning surfaces is not currently
                supported.
                """
            )



# -*- coding: utf-8 -*-
"""
Copied code from...
https://stackoverflow.com/questions/433162/
"""

from django import template
from django.conf import settings as django_settings

from simmate.configuration import settings as simmate_settings

register = template.Library()


@register.simple_tag
def django_setting(name):
    """
    Loads the value of a django setting

    Example use:
    ``` html
    {% load getsetting %}
    {% django_setting "DEBUG" %}
    ```
    """

    return getattr(django_settings, name, "")


@register.simple_tag
def simmate_setting(name):
    """
    Loads the value of a simmate setting

    Example use:
    ``` html
    {% load getsetting %}
    {% simmate_setting "website.require_login" %}
    ```
    """
    keys = name.split(".")
    assert len(keys) > 1  # ensure correct usage
    final_value = simmate_settings  # start with the full object
    for key in keys:  # recursively grab attributes
        final_value = getattr(final_value, key, "")
    return final_value



# -*- coding: utf-8 -*-

from simmate.apps.vasp.inputs import PBE_POTCAR_MAPPINGS
from simmate.apps.vasp.workflows.base import VaspWorkflow
from simmate.apps.vasp.workflows.diffusion import (
    NebAllPathsWorkflow,
    SinglePathWorkflow,
    VaspNebFromImagesWorkflow,
)

# -----------------------------------------------------------------------------

# BULK UNITCELL RELAXATION


class Relaxation__Vasp__WarrenLab(VaspWorkflow):
    functional = "PBE"
    potcar_mappings = PBE_POTCAR_MAPPINGS
    _incar = dict(
        ALGO="Fast",
        EDIFF=1e-06,
        ENCUT=520,  # TODO: set dynamically to be 1.3x highest elemental ENMAX
        IBRION=2,
        ICHARG=1,
        ISIF=3,
        ISPIN=2,
        ISYM=0,
        IVDW=12,
        LORBIT=11,
        LREAL="Auto",
        LWAVE=False,
        # MAGMOM = # TODO: set dynamically
        NELM=200,
        NELMIN=4,
        NSW=99,
        PREC="Accurate",
        ISMEAR=0,
        SIGMA=0.05,
        KSPACING=0.35,
        LMAXMIX=4,
    )
    error_handlers = []


# -----------------------------------------------------------------------------

# BULK UNITCELL STATIC ENERGY


class StaticEnergy__Vasp__WarrenLab(VaspWorkflow):
    functional = "PBE"
    potcar_mappings = PBE_POTCAR_MAPPINGS
    _incar = dict(
        ALGO="Fast",
        EDIFF=1e-06,
        ENCUT=520,  # TODO: set dynamically to be 1.3x highest elemental ENMAX
        IBRION=-1,
        ICHARG=1,
        ISIF=3,
        ISPIN=2,
        ISYM=0,
        IVDW=12,
        LORBIT=11,
        LREAL="Auto",
        LWAVE=False,
        # MAGMOM = # TODO: set dynamically
        NELM=200,
        NELMIN=4,
        NSW=0,
        PREC="Accurate",
        ISMEAR=0,
        SIGMA=0.05,
        KSPACING=0.35,
        LMAXMIX=4,
    )
    error_handlers = []


# -----------------------------------------------------------------------------

# ENDPOINT SUPERCELL RELAXATIONS


class Relaxation__Vasp__WarrenLabNebEndpoint(VaspWorkflow):
    functional = "PBE"
    potcar_mappings = PBE_POTCAR_MAPPINGS
    _incar = dict(
        ALGO="Fast",
        EDIFF=5e-05,
        EDIFFG=-0.01,
        ENCUT=520,  # TODO: set dynamically to be 1.3x highest elemental ENMAX
        IBRION=2,
        ICHARG=1,
        ISIF=2,
        ISPIN=2,
        ISYM=0,
        IVDW=12,
        LORBIT=11,
        LREAL="Auto",
        LWAVE=False,
        LCHARG=False,
        # MAGMOM = # TODO: set dynamically
        NELM=200,
        NSW=99,
        PREC="Accurate",
        ISMEAR=0,
        SIGMA=0.05,
        KSPACING=0.4,
        LMAXMIX=4,
    )
    error_handlers = []


# -----------------------------------------------------------------------------

# ENDPOINT SUPERCELL STATIC ENERGY


class StaticEnergy__Vasp__WarrenLabNebEndpoint(VaspWorkflow):
    functional = "PBE"
    potcar_mappings = PBE_POTCAR_MAPPINGS
    _incar = dict(
        ALGO="Fast",
        EDIFF=5e-05,
        EDIFFG=-0.01,
        ENCUT=520,  # TODO: set dynamically to be 1.3x highest elemental ENMAX
        IBRION=-1,
        ICHARG=1,
        ISIF=2,
        ISPIN=2,
        ISYM=0,
        IVDW=12,
        LORBIT=11,
        LREAL="Auto",
        LWAVE=False,
        LCHARG=False,
        # MAGMOM = # TODO: set dynamically
        NELM=200,
        NSW=0,
        PREC="Accurate",
        ISMEAR=0,
        SIGMA=0.05,
        KSPACING=0.4,
        LMAXMIX=4,
    )
    error_handlers = []


# -----------------------------------------------------------------------------

# NEB FROM IMAGES


class Diffusion__Vasp__WarrenLabCiNebFromImages(VaspNebFromImagesWorkflow):
    functional = "PBE"
    potcar_mappings = PBE_POTCAR_MAPPINGS
    _incar = dict(
        ALGO="Fast",
        EDIFF=5e-05,
        EDIFFG=-0.01,
        ENCUT=520,  # TODO: set dynamically to be 1.3x highest elemental ENMAX
        IBRION=3,
        ICHARG=1,
        ISIF=2,
        ISPIN=2,
        ISYM=0,
        IVDW=12,
        LORBIT=11,
        LREAL="Auto",
        LWAVE=False,
        LCHARG=False,
        # MAGMOM = # TODO: set dynamically
        NELM=200,
        NSW=99,
        PREC="Accurate",
        ISMEAR=0,
        SIGMA=0.05,
        KSPACING=0.4,
        LMAXMIX=4,
        NIMAGES__auto=True,  # set by simmate
        LCLIMB=True,
        SPRING=-5,
        POTIM=0,
        IOPT=1,
    )
    error_handlers = []


# -----------------------------------------------------------------------------

# SINGLE PATH NEB


class Diffusion__Vasp__WarrenLabNebSinglePath(SinglePathWorkflow):
    endpoint_relaxation_workflow = Relaxation__Vasp__WarrenLabNebEndpoint
    endpoint_energy_workflow = StaticEnergy__Vasp__WarrenLabNebEndpoint
    from_images_workflow = Diffusion__Vasp__WarrenLabCiNebFromImages


# -----------------------------------------------------------------------------

# ALL PATHS NEB


class Diffusion__Vasp__NebAllPathsWarrenLab(NebAllPathsWorkflow):
    bulk_relaxation_workflow = Relaxation__Vasp__WarrenLab
    bulk_static_energy_workflow = StaticEnergy__Vasp__WarrenLab
    single_path_workflow = Diffusion__Vasp__WarrenLabNebSinglePath


# -----------------------------------------------------------------------------



# -*- coding: utf-8 -*-

from simmate.apps.warren_lab.workflows.relaxation.pbe_metal import (
    Relaxation__Vasp__WarrenLabPbeMetal,
)
from simmate.apps.warren_lab.workflows.static_energy.pbe import PBE_STATIC_SETTINGS


class StaticEnergy__Vasp__WarrenLabPbeMetal(Relaxation__Vasp__WarrenLabPbeMetal):
    """
    Performs a static energy calculation based on the settings for Warren Lab
    Metal relaxations using the PBE functional.
    """

    _incar_updates = PBE_STATIC_SETTINGS



# -*- coding: utf-8 -*-

from simmate.apps.materials_project.workflows.relaxation.mit import (
    Relaxation__Vasp__Mit,
)
from simmate.apps.vasp.error_handlers import Frozen, Unconverged, Walltime
from simmate.apps.vasp.workflows.diffusion import VaspNebFromImagesWorkflow


class Diffusion__Vasp__NebFromImagesMit(
    VaspNebFromImagesWorkflow, Relaxation__Vasp__Mit
):
    """
    This task is a reimplementation of pymatgen's
    [MITNEBSet](https://pymatgen.org/pymatgen.io.vasp.sets.html#pymatgen.io.vasp.sets.MITNEBSet).

    Runs a NEB relaxation on a list of structures (aka images) using MIT Project
    settings. The lattice remains fixed and symmetry is turned off for this
    relaxation.

    You shouldn't use this workflow directly, but instead use the higher-level
    NEB workflows (e.g. diffusion/neb_all_paths or diffusion/neb_from_endpoints),
    which call this workflow for you.
    """

    _incar_updates = dict(
        IBRION=1,
        ISYM=0,
        LCHARG=False,
        IMAGES__auto=True,
    )

    # Because of NEB's unique folder structure, many error handlers are broken
    # and cannot be used. We go through the list that is broken and remove them
    error_handlers = [
        handler
        for handler in Relaxation__Vasp__Mit.error_handlers
        if type(handler) not in [Walltime, Unconverged, Frozen]
    ]



# -*- coding: utf-8 -*-

"""
Building a workflow in Simmate can sometimes involve using custom database
tables, so before you start editting this file, make sure you have gone through
and editted the `models.py` file.

Below is an example of a simple VaspWorkflow, which is used to run a single VASP
calculation.

As your need grow, you may need more complex Workflows. For complex workflows that
require several calculations, other python logic, and parallelization, make
sure you read through the documentation for 
[the Workflow class](https://jacksund.github.io/simmate/full_guides/workflows/overview/) 
as well as relevent subclasses like the
[S3Workflow](https://jacksund.github.io/simmate/full_guides/workflows/s3_workflows/)
or 
[VaspWorkflow](https://jacksund.github.io/simmate/full_guides/workflows/third_party_software/vasp/overview/)
"""

from simmate.engine import Workflow, workflow
from simmate.toolkit import Structure

# Import our tables from the other files.
from .models import MyCustomTable2

# -----------------------------------------------------------------------------
# Make sure to list out all workflows that you want registered
# -----------------------------------------------------------------------------

__all__ = [
    "add",
    "Example__Python__MyExample",
]

# -----------------------------------------------------------------------------
# Our first example shows off the basics
# -----------------------------------------------------------------------------


@workflow
def add(x, y, **kwargs):
    """
    A basic workflow that adds two numbers together
    """
    return x + y


class Example__Python__MyExample(Workflow):
    """
    A basic workflow that returns a dictionary, which then updates a database table.

    Once ran, note how the structure and inputs are saved at the start of the
    calculation. Plus the outputs are saved once it finishes
    """

    database_table = MyCustomTable2  # tells the workflow to use this table

    @staticmethod
    def run_config(
        input_01: float,
        input_02: any,
        structure: Structure,
        **kwargs,
    ):
        # This is a boring workflow because it just saves the input values
        # to our table. You can get much more creative with workflows though
        return {
            "output_01": input_01 * 100,
            "output_02": False,
        }



# -*- coding: utf-8 -*-

from simmate.engine.execution.cluster import LocalCluster, SlurmCluster


def start_cluster(
    nworkers: int,
    cluster_type: str = "local",
    continuous: bool = False,
):
    """
    Utilitiy that helps set up common cluster types with a specific number of
    workers and optionally run a single-submit of workers
    """
    if cluster_type == "local":
        cluster = LocalCluster
    elif cluster_type == "slurm":
        cluster = SlurmCluster
    else:
        raise Exception(f"Unknown cluster type {cluster_type}. Choose local or slurm.")

    if continuous:
        cluster.start_cluster(nworkers)
    else:
        jobs = cluster.submit_jobs(nworkers)
        if cluster_type == "local":
            cluster.wait_for_jobs(jobs)



# -*- coding: utf-8 -*-

from simmate.apps.warren_lab.workflows.relaxation.hse import (
    Relaxation__Vasp__WarrenLabHse,
)

HSE_STATIC_SETTINGS = dict(
    IBRION=-1,  # (optional) locks everything between ionic steps
    NSW=0,  # this is the main static energy setting
    #            LAECHG=True, # currently only set in population analysis
    LCHARG=True,
    LORBIT=11,
    LVHAR=True,
    LWAVE=True,
    # The only difference between this and PBE is that we don't want to change
    # the ALGO tag to "Normal" for HSE. It's recommended to use ALGO="All" for
    # insulators or ALGO="Damped" for small bandgap/metals
)


class StaticEnergy__Vasp__WarrenLabHse(Relaxation__Vasp__WarrenLabHse):
    """
    Performs a static energy calculation based on the settings for Warren Lab
    HSE functional relaxation.
    """

    _incar_updates = HSE_STATIC_SETTINGS



# -*- coding: utf-8 -*-

from simmate.database.base_data_types import Calculation, table_column


class VariableNsitesCompositionSearch(Calculation):
    class Meta:
        app_label = "workflows"

    # OPTIMIZE: This is really just a copy/paste of the fixed-composition
    # workflow, where we aren't using the table directly in the search and
    # skips making steadystate_source table entries. It simply stores input
    # parameters.
    composition = table_column.CharField(max_length=50, null=True, blank=True)
    subworkflow_name = table_column.CharField(max_length=200, null=True, blank=True)
    subworkflow_kwargs = table_column.JSONField(default=dict, null=True, blank=True)
    fitness_field = table_column.CharField(max_length=200, null=True, blank=True)
    min_structures_exact = table_column.IntegerField(null=True, blank=True)
    max_structures = table_column.IntegerField(null=True, blank=True)
    best_survival_cutoff = table_column.IntegerField(null=True, blank=True)
    nfirst_generation = table_column.IntegerField(null=True, blank=True)
    nsteadystate = table_column.IntegerField(null=True, blank=True)
    convergence_cutoff = table_column.FloatField(null=True, blank=True)
    selector_name = table_column.CharField(max_length=200, null=True, blank=True)
    selector_kwargs = table_column.JSONField(default=dict, null=True, blank=True)
    validator_name = table_column.CharField(max_length=200, null=True, blank=True)
    validator_kwargs = table_column.JSONField(default=dict, null=True, blank=True)
    sleep_step = table_column.FloatField(null=True, blank=True)
    expected_structure = table_column.JSONField(default=dict)



# -*- coding: utf-8 -*-

import logging

from simmate.toolkit import Structure


class Transformation:
    """
    Transformations (aka "mutations") alter an input structure (or structures)
    in some way. For now we assume all transformations are for toolkit
    `Structure` objects

    Different transformation can take a different number of structures as inputs
    and also potentially return more than one structure. Therefore, pay close
    attention to setting `io_scale` and `ninput`

    ## Example use

    When writing a custom transformation class, you can use the following example:

    ``` python
    class SubClass(Transformation):
        io_scale = 'one_to_one'
        ninput = 1
        allow_parallel = True
        def apply_transformation(self, structure):
            # do some transform on the input
            return new_structure

    # example use
    t = SubClass()
    new_structure = t.apply_transformation(structure)
    ```

    This is a refactored version of pymatgen's
    [Transformation](https://pymatgen.org/pymatgen.transformations.transformation_abc.html)
    class because they do not support many_to_one or many_to_many
    (see pymatgen.transformations.transformation_abc)
    """

    io_type: str = "structure"
    """
    Should I do StructureTransformation, LatticeTransformation, and
    SiteTransformation subclasses? Here, io_type could be either
    'structure', 'lattice', or 'site' as to indicate the input.
    """
    #!!! For now, I have all transformations as StructureTransformation
    #!!! so I assume io_type here. This may change in the future.

    allow_parallel: bool = False
    """
    Whether or not this transformation can be done in parallel.
    
    In pymatgen, this parameter is named `use_multiprocessing`
    """

    @property
    def io_scale(self) -> str:
        """
        The input/output type for the class.
        This should be one of the following choices:
            - one_to_one
            - one_to_many
            - many_to_one
            - many_to_many # I have no examples of this yet
        """
        raise NotImplementedError(
            "Make sure you set a `io_scale` for your Transformation class"
        )

    @property
    def ninput(self):
        """
        Number of inputs required. For example, some transformations require
        two structures to be input. If one_to_* we know ninput = 1.
        """
        raise NotImplementedError(
            "Make sure you set `ninput` for your Transformation class"
        )

    def apply_transformation(self, structure: Structure) -> Structure:
        """
        This method carries out the transformation. The typing shown above
        is an example of a `one_to_one` transformation that accepts a
        single structure and returns a single one.
        """
        raise NotImplementedError(
            "Make sure you define a new `apply_transformation` for "
            "your Transformation class"
        )

    @classmethod
    @property
    def name(cls):
        """
        A nice string name for the transformation. By default it just returns
        the name of this class.
        """
        return cls.__name__

    def apply_from_database_and_selector(
        self,
        selector,
        datatable,
        select_kwargs: dict = {},
        max_attempts: int = 100,
        **kwargs,  # for apply_transformation_with_validation
    ):
        logging.info(f"Creating a transformed structure with '{self.name}'")
        logging.info(f"Parent(s) will be selected using '{selector.name}'")

        # grab parent structures using the selection method
        parent_ids, parent_structures = selector.select_from_datatable(
            nselect=self.ninput,
            datatable=datatable,
            **select_kwargs,
        )

        logging.info(f"Selected parents: {parent_ids}")

        # Until we get a new valid structure (or run out of attempts), keep trying
        # with our given source. Assume we don't have a valid structure until
        # proven otherwise
        new_structure = False
        attempt = 0
        while not new_structure and attempt <= max_attempts:
            # add an attempt
            attempt += 1

            # make a new structure
            new_structure = self.apply_transformation_with_validation(
                parent_structures,
                **kwargs,
            )

        # see if we got a structure or if we hit the max attempts and there's
        # a serious problem!
        if not new_structure:
            logging.warning(
                f"Failed to create a structure after {max_attempts} different"
                "parent combinations. Giving up."
            )
            return False

        # Otherwise we were successful
        logging.info("Creation Successful.")

        return parent_ids, new_structure

    def apply_transformation_with_validation(
        self,
        structures,  # either a structure or list of structures. Depends on ninput.
        validators=[],
        max_attempts=100,
    ):
        # Until we get a new valid structure (or run out of attempts), keep trying
        # with our given source. Assume we don't have a valid structure until
        # proven otherwise
        new_structure = False
        attempt = 0
        while not new_structure and attempt <= max_attempts:
            # add an attempt
            attempt += 1

            # make a new structure
            new_structure = self.apply_transformation(structures)

            # check to see if the structure passes all validation checks.
            if new_structure:
                for validator in validators:
                    is_valid = validator.check_structure(new_structure)

                    if not is_valid:
                        # if it is not unique, we can throw away the structure and
                        # try the loop again.
                        logging.debug(
                            "Generated structure failed validation by "
                            f"{validator.name}. Trying again."
                        )
                        new_structure = None

                        # one failed validation is enough to stop. There is no
                        # need to test the other validation methods.
                        break

        # see if we got a structure or if we hit the max attempts and there's
        # a serious problem!
        if not new_structure:
            logging.warning(
                "Failed to create a structure from input after "
                f"{max_attempts} attempts"
            )
            return False

        # return the structure and its parents
        return new_structure



# -*- coding: utf-8 -*-

from .aflow import AflowPrototype, AflowStructure
from .cod import CodStructure
from .jarvis import JarvisStructure
from .materials_project import MatprojStructure
from .oqmd import OqmdStructure
from .utilities import load_default_sqlite3_build, load_remote_archives



# -*- coding: utf-8 -*-

from django_filters import rest_framework as django_api_filters
from scipy.constants import Avogadro

from simmate.database.base_data_types import DatabaseTable, Spacegroup, table_column
from simmate.toolkit import Structure as ToolkitStructure
from simmate.utilities import get_chemical_subsystems


class Structure(DatabaseTable):
    class Meta:
        abstract = True

    exclude_from_summary = ["structure", "elements"]

    archive_fields = ["structure"]

    api_filters = dict(
        nsites=["range"],
        nelements=["range"],
        # elements=["contains"],
        density=["range"],
        density_atomic=["range"],
        volume=["range"],
        volume_molar=["range"],
        formula_full=["exact"],
        formula_reduced=["exact"],
        formula_anonymous=["exact"],
        spacegroup__number=["exact"],
        spacegroup__symbol=["exact"],
        spacegroup__crystal_system=["exact"],
        spacegroup__point_group=["exact"],
        # Whether to include subsystems of the given `chemical_system`. For
        # example, the subsystems of Y-C-F would be Y, C, F, Y-C, Y-F, etc..
        include_subsystems=django_api_filters.BooleanFilter(
            field_name="include_subsystems",
            label="Include chemical subsystems in results?",
            method="skip_filter",
        ),
        # TODO: Supra-systems would include all the elements listed AND more. For example,
        # searching Y-C-F would also return Y-C-F-Br, Y-Sc-C-F, etc.
        # include_suprasystems = forms.BooleanField(label="Include Subsytems", required=False)
        # The chemical system of the structure (e.g. "Y-C-F" or "Na-Cl")
        chemical_system=django_api_filters.CharFilter(
            method="filter_chemical_system",
        ),
    )

    structure = table_column.TextField(blank=True, null=True)
    """
    The core structure information, which is written to a string and in a 
    compressed format using the `from_toolkit` method. To get back to our toolkit
    structure object, use the `to_toolkit` method.
    """

    nsites = table_column.IntegerField(blank=True, null=True)
    """
    The total number of sites in the unitcell. (e.g. Y2CF2 has 5 sites)
    """

    nelements = table_column.IntegerField(blank=True, null=True)
    """
    The total number of unique elements. (e.g. Y2CF2 has 3 elements)
    """

    elements = table_column.JSONField(blank=True, null=True)
    """
    List of elements in the structure (ex: ["Y", "C", "F"])
    """

    chemical_system = table_column.CharField(max_length=25, blank=True, null=True)
    """
    the base chemical system (ex: "Y-C-F")
    
    Note: be careful when searching for elements! Running chemical_system__contains="C"
    on this field won't do what you expect -- because it will return structures
    containing Ca, Cs, Ce, Cl, and so on. If you want to search for structures
    that contain a specific element, use elements__contains='"C"' instead. The
    odd use of quotes '"C"' is required here!
    """

    density = table_column.FloatField(blank=True, null=True)
    """
    The density of the crystal in g/cm^3
    """

    density_atomic = table_column.FloatField(blank=True, null=True)
    """
    The density of atoms in the crystal in atoms/Angstom^3
    """

    volume = table_column.FloatField(blank=True, null=True)
    """
    The volume of the unitcell in Angstom^3
    
    Note: in most cases, `volume_molar` should be used instead! This is because
    volumne is highly dependent on the symmetry and the arbitray unitcell. If 
    you are truly after small volumes of the unitcell, it is likely you really 
    just want to search by spacegroup.
    """

    volume_molar = table_column.FloatField(blank=True, null=True)
    """
    The molar volume of the crystal in cm^3/mol
    """

    # The composition of the structure formatted in various ways
    # BUG: The max length here is overkill because there are many structures
    # with 8+ elements and disordered formula (e.g. "Ca2.103 N0.98")

    formula_full = table_column.CharField(
        max_length=50,
        blank=True,
        null=True,
    )
    """
    The chemical formula with elements sorted by electronegativity (ex: Li4 Fe4 P4 O16)
    """

    formula_reduced = table_column.CharField(
        max_length=50,
        blank=True,
        null=True,
    )
    """
    The reduced chemical formula. (ex: Li4Fe4P4O16 --> LiFePO4)
    """

    formula_anonymous = table_column.CharField(
        max_length=50,
        blank=True,
        null=True,
    )
    """
    An anonymized formula. Unique species are arranged in ordering of 
    amounts and assigned ascending alphabets. Useful for prototyping formulas. 
    For example, all stoichiometric perovskites have anonymized_formula ABC3.
    """

    spacegroup = table_column.ForeignKey(
        Spacegroup,
        on_delete=table_column.PROTECT,
        blank=True,
        null=True,
    )
    """
    Spacegroup information. Points to a separate database table that has additional
    columns:
    `simmate.database.base_data_types.symmetry.Spacegroup`
    """

    # The AFLOW prototype that this structure maps to.
    # TODO: this will be a relationship in the future
    # prototype = table_column.CharField(max_length=50, blank=True, null=True)

    # NOTE: extra fields for the Lattice and Sites are intentionally left out
    # in order to save on overall database size. Things such as...
    #   Lattice: matrix and then... a, b, c, alpha, beta, gamma, volume
    #   Sites: abc, xyz, properties, species/element, occupancy
    # shouldn't be queried directly. If you'd like to sort structures by these
    # criteria, you can still do this in python and pandas! Just not at the
    # SQL level

    def filter_chemical_system(self, queryset, name, value):
        # name/value here are the key/value pair for chemical system

        # Grab the "include_subsystems" field from the filter form. Note, this
        # value will be given as a string which we convert to a python boolean
        include_subsystems = self.data.dict().get("include_subsystems", "false")
        include_subsystems = True if include_subsystems == "true" else False

        # TODO:
        # Make sure that the chemical system is made of valid elements and
        # separated by hyphens

        # check if the user wants subsystems included (This will be True or False)
        if include_subsystems:
            systems_cleaned = get_chemical_subsystems(value)

        # otherwise just clean the single system
        else:
            # Convert the system to a list of elements
            systems_cleaned = value.split("-")
            # now recombine the list back into alphabetical order
            systems_cleaned = ["-".join(sorted(systems_cleaned))]
            # NOTE: we call this "systems_cleaned" and put it in a list so
            # that our other methods don't have to deal with multiple cases
            # when running a django filter.

        filtered_queryset = queryset.filter(chemical_system__in=systems_cleaned)

        # now return the cleaned value. Note that this is now a list of
        # chemical systems, where all elements are in alphabetical order.
        return filtered_queryset

    @classmethod
    def _from_toolkit(
        cls,
        structure: ToolkitStructure | str = None,
        as_dict: bool = False,
        **kwargs,
    ):
        # if there isn't a structure, nothing is to be done.
        if not structure:
            return kwargs if as_dict else cls(**kwargs)

        if isinstance(structure, str):
            structure = ToolkitStructure.from_database_string(structure)

        # BUG: This is an old line and I can't remember why I have it. Once I
        # have implemented more unittests, consider deleting. This method is
        # ment to convert to a ToolkitStructure, but the structure should
        # already be in this format...
        structure = ToolkitStructure.from_dynamic(structure)

        # OPTIMIZE: I currently store files as poscar strings for ordered structures
        # and as CIFs for disordered structures. Both of this include excess information
        # that slightly inflates file size, so I will be making a new string format in
        # the future. This will largely be based off the POSCAR format, but will
        # account for disordered structures and all limit repeated data (such as the
        # header line, "direct", listing each element/composition, etc.).
        storage_format = "POSCAR" if structure.is_ordered else "CIF"

        # OPTIMIZE
        # This attempts to match the structure to an AFLOW prototype and it is
        # by far the slowest step of loading structures to the database. Try
        # to optimize this in the future.
        # from pymatgen.analysis.prototypes import AflowPrototypeMatcher
        # prototype = AflowPrototypeMatcher().get_prototypes(structure)
        # prototype_name = prototype[0]["tags"]["mineral"] if prototype else None
        # Alternatively, add as a method to the table, similar to
        # the "update_all_stabilities" for thermodynamics

        # Given a pymatgen structure object, this will return a database structure
        # object, but will NOT save it to the database yet. The kwargs input
        # is only if you inherit from this class and add extra fields.
        structure_dict = dict(
            structure=structure.to(fmt=storage_format),
            nsites=structure.num_sites,
            nelements=len(structure.composition),
            elements=[str(e) for e in structure.composition.elements],
            chemical_system=structure.composition.chemical_system,
            density=float(structure.density),
            density_atomic=structure.num_sites / structure.volume,
            volume=structure.volume,
            # 1e-27 is to convert from cubic angstroms to Liter and then 1e3 to
            # mL. Therefore this value is in mL/mol
            # OPTIMIZE: move this to a class method
            volume_molar=(structure.volume / structure.num_sites)
            * Avogadro
            * 1e-27
            * 1e3,
            # OPTIMIZE SPACEGROUP INFO
            spacegroup_id=structure.get_space_group_info(
                symprec=0.1,
                # angle_tolerance=5.0,
            )[1],
            formula_full=structure.composition.formula,
            formula_reduced=structure.composition.reduced_formula,
            formula_anonymous=structure.composition.anonymized_formula,
            # prototype=prototype_name,
            **kwargs,  # this allows subclasses to add fields with ease
        )
        # If as_dict is false, we build this into an Object. Otherwise, just
        # return the dictionary
        return structure_dict if as_dict else cls(**structure_dict)

    def to_toolkit(self) -> ToolkitStructure:
        """
        Converts the database object to toolkit Structure object.
        """
        return ToolkitStructure.from_database_object(self)



# -*- coding: utf-8 -*-

from simmate.apps.warren_lab.workflows.relaxation.pbe import (
    Relaxation__Vasp__WarrenLabPbe,
)


class Relaxation__Vasp__WarrenLabPbeMetal(Relaxation__Vasp__WarrenLabPbe):
    """
    Runs a VASP relaxation calculation using Warren Lab PBE functional settings
    with some adjusted settings to accomodate metals.
    """

    description_doc_short = "Warren Lab presets for metals"

    _incar_updates = dict(
        ISMEAR=1,
        SIGMA=0.2,
        KSPACING=0.3,
    )



from simmate.apps.vasp.inputs import PBE_POTCAR_MAPPINGS
from simmate.apps.vasp.workflows.base import VaspWorkflow


class Relaxation__Vasp__ClusterHighQ(VaspWorkflow):
    functional = "PBE"
    potcar_mappings = PBE_POTCAR_MAPPINGS

    _incar = dict(
        IBRION=2,
        KSPACING=0.4,
        EDIFFG=-0.02,
        IVDW=11,
        PREC="Accurate",
        ISMEAR=-5,
        ALGO="Fast",
        EDIFF=1e-06,
        ENCUT=500,
        NELM=100,
        NSW=99,
        SIGMA=0.05,
        LASPH=True,
        LORBIT=11,
        multiple_keywords__smart_ldau=dict(
            LDAU__auto=True,
            LDAUTYPE=2,
            LDAUPRINT=2,
            LDAUJ={"C": {"Sc": 0}},
            LDAUL={"C": {"Sc": 2}},
            LDAUU={"C": {"Sc": 3.0}},
        ),
    )
